# Seung lab branch of easyviewer

import re
from collections import OrderedDict
from typing import Dict, List, Optional, Tuple, Union
from warnings import warn

from numpy import integer, issubdtype, uint64, vstack

from ...site_utils import SEUNGLAB_NAMES, neuroglancer_url
from . import nglite as neuroglancer
from . import utils
from .base import SEGMENTATION_LAYER_TYPES, EasyViewerBase

SEUNGLAB_NAME = SEUNGLAB_NAMES[0]


class EasyViewerSeunglab(neuroglancer.UnsynchronizedViewer, EasyViewerBase):
    def __init__(self, **kwargs):
        super(neuroglancer.UnsynchronizedViewer, self).__init__(**kwargs)
        super(EasyViewerBase, self).__init__(**kwargs)

    def load_url(self, url) -> None:
        "Parse a neuroglancer state based on URL and load it into the state"
        state = neuroglancer.parse_url(url)
        self.set_state(state)

    def _ImageLayer(self, source, **kwargs):
        return neuroglancer.ImageLayer(source=source, **kwargs)

    def _AnnotationLayer(self, **kwargs):
        return neuroglancer.AnnotationLayer(
            **kwargs,
        )

    def _SegmentationLayer(self, source, **kwargs):
        if re.search(r"^graphene://", source) is not None:
            source = utils.parse_graphene_header(source, SEUNGLAB_NAME)
            return neuroglancer.ChunkedgraphSegmentationLayer(source=source, **kwargs)
        elif re.search(r"^precomputed://", source) is not None:
            return neuroglancer.SegmentationLayer(source=source, **kwargs)
        else:
            raise ValueError("Source must be either graphene:// or precomputed://")

    def set_resolution(self, resolution) -> None:
        with self.txn() as s:
            s.voxel_size = resolution

    def set_state_server(self, state_server) -> None:
        with self.txn() as s:
            s._json_data["jsonStateServer"] = state_server

    def add_segmentation_layer(self, layer_name, source, **kwargs):
        """Add segmentation layer to viewer instance.

        Attributes:
            layer_name (str): name of layer to be displayed in neuroglancer ui.
            source (str): source of neuroglancer segment layer
        """
        if type(source) is not str:
            source = source[0]
            warn("Only using first source in list for seung-lab segmentation layer")
        super().add_segmentation_layer(layer_name, source, **kwargs)

    def append_source_to_segmentation_layer(self, layer_name, source):
        warn("Additional segment source is not supported in Seung-lab viewer")

    def add_skeleton_source(self, layer_name, source, shader_text=None):
        warn("Skeleton sources are not supported in Seung-lab viewer")

    def set_skeleton_shader(self, layer_name, shader_text=None):
        warn("Skeleton shaders are not supported in Seung-lab viewer")

    def add_annotation_layer(
        self,
        layer_name=None,
        color=None,
        linked_segmentation_layer=None,
        filter_by_segmentation=False,
        brackets_show_segmentation=True,
        selection_shows_segmentation=True,
        tags=None,
    ) -> None:
        if layer_name is None:
            layer_name = "annos"
        if layer_name in [l.name for l in self.state.layers]:
            raise ValueError("Layer name already exists")
        if linked_segmentation_layer is None:
            filter_by_segmentation = None

        with self.txn() as s:
            new_layer = self._AnnotationLayer(
                linked_segmentation_layer=linked_segmentation_layer,
                filter_by_segmentation=filter_by_segmentation,
                brackets_show_segmentation=brackets_show_segmentation,
                selection_shows_segmentation=selection_shows_segmentation,
            )
            s.layers.append(name=layer_name, layer=new_layer)
            if color is not None:
                s.layers[layer_name].annotationColor = utils.parse_color(color)

        if tags is not None:
            self.add_annotation_tags(layer_name=layer_name, tags=tags)

    def _convert_annotations(self, annotations: List) -> List:
        """Pass through annotations, currently defaulting to seung lab format already"""
        return annotations

    def add_annotation_tags(self, layer_name, tags):
        if layer_name not in self.layer_names:
            raise ValueError("Layer is not an annotation layer")
        tag_list = [
            OrderedDict({"id": tag_id + 1, "label": label})
            for tag_id, label in enumerate(tags)
        ]
        with self.txn() as s:
            s.layers[layer_name].annotationTags = tag_list

    def as_url(
        self,
        prefix: Optional[str] = None,
        as_html: Optional[bool] = False,
        link_text: Optional[str] = "Neuroglancer Link",
    ) -> str:
        prefix = neuroglancer_url(
            prefix, SEUNGLAB_NAME
        )  # 'seunglab' hard-coded because of file.
        ngl_url = neuroglancer.to_url(self.state, prefix=prefix)
        if as_html:
            return f'<a href="{ngl_url}" target="_blank">{link_text}</a>'
        else:
            return ngl_url

    def select_annotation(self, layer_name, anno_id):
        if layer_name in self.layer_names:
            with self.txn() as s:
                s.layers[layer_name]._json_data["selectedAnnotation"] = id
        self.set_selected_layer(layer_name)

    def set_selected_layer(self, layer_name):
        with self.txn() as s:
            s.selectedLayer.layer = layer_name

    def add_selected_objects(
        self,
        segmentation_layer: str,
        oids: List[int],
        colors: Optional[Union[List, Dict]] = None,
    ) -> None:
        if issubdtype(type(oids), integer):
            oids = [oids]

        with self.txn() as s:
            for oid in oids:
                s.layers[segmentation_layer].segments.add(uint64(oid))
            if s.layers[segmentation_layer].type == "segmentation_with_graph":
                s.layers[segmentation_layer].segmentQuery = ", ".join(
                    [str(x) for x in oids]
                )
        if colors is not None:
            if isinstance(colors, dict):
                self.assign_colors(segmentation_layer, colors)
            elif len(colors) == len(oids):
                seg_colors = {str(oid): clr for oid, clr in zip(oids, colors)}
                self.assign_colors(segmentation_layer, seg_colors)

    def assign_colors(self, layer_name, seg_colors):
        """Assign colors to root ids in a segmentation layer

        Parameters
        ----------
        layer_name : str,
            Segmentation layer name
        seg_colors : dict
            dict with root ids as keys and colors as values.
        """
        with self.txn() as s:
            if seg_colors is not None:
                seg_colors = {
                    str(oid): utils.parse_color(k)
                    for oid, k in seg_colors.items()
                    if k is not None
                }
                s.layers[layer_name].segmentColors = seg_colors

    def set_view_options(
        self,
        show_slices: Optional[bool] = None,
        layout: Optional[str] = None,
        show_axis_lines: Optional[bool] = None,
        show_scale_bar: Optional[bool] = None,
        orthographic: Optional[bool] = None,
        position: Optional[Tuple[float]] = None,
        zoom_image: Optional[float] = None,
        zoom_3d: Optional[float] = None,
        background_color: Optional[Tuple[float]] = None,
    ) -> None:
        """Set options relating to the neuroglancer view state. Only changes the values of the parameters provided.

        Parameters
        ----------
        show_slices : bool, optional
            Show slice cutout in the 3d view, by default None
        layout : str, optional
            Change the layout type ('xy', 'yz', 'zx', '3d', 'xy-3d', 'yz-3d', 'zx-3d', or '4panel'), by default None
        show_axis_lines : bool, optional
            Show the red/blue/green lines indicating the axis directions, by default None
        show_scale_bar : bool, optional
            Controls showing of scale bar, by default None
        orthographic : bool, optional
            Controls whether the 3d perspective view is orthographic or not, by default None
        position : list, optional
            Sets the location of center point of the view in Neuroglancer coordinates, by default None
        zoom_image : int, optional
            Sets the zoom factor for the imagery, by default None
        zoom_3d : int, optional
            Sets the zoom factor for the 3d view, by default None
        background_color : list or str, optional
            hex, rgb, or named color for the background of the 3d viewer, by default None
        """
        with self.txn() as s:
            if show_slices is not None:
                s.showSlices = show_slices
            if layout is not None:
                s.layout.type = layout
            if show_axis_lines is not None:
                s.show_axis_lines = show_axis_lines
            if show_scale_bar is not None:
                s.show_scale_bar = show_scale_bar
            if orthographic is not None:
                s.layout.orthographic_projection = orthographic
            if position is not None:
                s.position.voxelCoordinates = position
            if zoom_image is not None:
                s.navigation.zoomFactor = zoom_image
            if zoom_3d is not None:
                s.perspectiveZoom = zoom_3d
            if background_color is not None:
                s.perspectiveViewBackgroundColor = utils.parse_color(background_color)

    def set_segmentation_view_options(
        self,
        layer_name: str,
        alpha_selected: Optional[float] = None,
        alpha_3d: Optional[float] = None,
        alpha_unselected: Optional[float] = None,
        **kwargs,
    ):
        if self.state.layers[layer_name].type not in SEGMENTATION_LAYER_TYPES:
            raise ValueError("Layer is not a segmentation layer")
        with self.txn() as s:
            l = s.layers[layer_name]
            if alpha_selected is not None:
                l.selectedAlpha = alpha_selected
            if alpha_3d is not None:
                l.objectAlpha = alpha_3d
            if alpha_unselected is not None:
                l.notSelectedAlpha = alpha_unselected

    def set_timestamp(
        self,
        layer_name,
        timestamp: Optional[int] = None,
    ):
        """Set timestamp of a segmentation layer

        Parameters
        ----------
        layer_name : str
            Name of a segmentation layer
        timestamp : float, optional
            Timestamp in unix epoch time (e.g. `time.time.now()` in python), by default None
        """
        if self.state.layers[layer_name].type != "segmentation_with_graph":
            return
        with self.txn() as s:
            l = s.layers[layer_name]
            if timestamp is not None:
                l.timestamp = int(timestamp)
            else:
                l.timestamp = None

    def set_multicut_points(
        self,
        layer_name,
        seg_id,
        points_red,
        points_blue,
        supervoxels_red=None,
        supervoxels_blue=None,
        focus=True,
    ):
        """Configures multicut points in the neuroglancer state. Note that points need to be in mesh units (e.g. nanometers), not voxels!

        Parameters
        ----------
        layer_name : str
            Segmentation layer name
        seg_id : np.uint64
            Segmentation id of the object in question
        points_red : np.array
            Nx3 array of locations in voxel space for side 1 of the cut.
        points_blue : np.array
            Mx3 array of locations in voxel space for side 2 of the cut.
        supervoxels_red : np.array or None, optional
            N-length array of supervoxel ids associated with locations in points_red or None.
            If None, supervoxels lookup occurs based on the mesh. By default None
        supervoxels_blue : np.array or None, optional
            M-length array of supervoxel ids associated with locations in points_blue or None.
            If None, supervoxels lookup occurs based on the mesh. By default None
        focus : bool, optional
            If True, makes the layer and graph tool focused. By default True
        """

        def _multicut_annotation(pt, oid, sv_id):
            if sv_id is None:
                sv_id = oid
            return neuroglancer.annotation.point_annotation(
                pt, description=str(sv_id), linked_segmentation=[sv_id, oid]
            )

        if supervoxels_red is None:
            supervoxels_red = [None for x in points_red]
        if supervoxels_blue is None:
            supervoxels_blue = [None for x in points_blue]

        if self.state.voxel_size is None:
            voxel_size = [4, 4, 40]
        else:
            voxel_size = self.state.voxel_size

        annos_red = neuroglancer.annotationHolder()
        for pt, sv_id in zip(points_red, supervoxels_red):
            annos_red.annotations.append(
                _multicut_annotation(pt * voxel_size, seg_id, sv_id)
            )

        annos_blue = neuroglancer.annotationHolder()
        for pt, sv_id in zip(points_blue, supervoxels_blue):
            annos_blue.annotations.append(
                _multicut_annotation(pt * voxel_size, seg_id, sv_id)
            )

        self.add_selected_objects(layer_name, [seg_id])

        with self.txn() as s:
            l = s.layers[layer_name]
            l.tab = "graph"
            l.graphOperationMarker.append(annos_red)
            l.graphOperationMarker.append(annos_blue)

        if focus:
            self.set_selected_layer(layer_name)
            ctr_pt = vstack([points_red, points_blue]).mean(axis=0)
            self.set_view_options(position=ctr_pt, zoom_3d=100)

    @staticmethod
    def point_annotation(
        point,
        id=None,
        description=None,
        linked_segmentation=None,
        tag_ids=None,
        **kwargs,
    ):
        return neuroglancer.point_annotation(
            point=point,
            id=id,
            description=description,
            linked_segmentation=linked_segmentation,
            tag_ids=tag_ids,
        )

    @staticmethod
    def line_annotation(
        pointA,
        pointB,
        id=None,
        description=None,
        linked_segmentation=None,
        tag_ids=None,
    ):
        return neuroglancer.line_annotation(
            a=pointA,
            b=pointB,
            id=id,
            description=description,
            linked_segmentation=linked_segmentation,
            tag_ids=tag_ids,
        )

    @staticmethod
    def ellipsoid_annotation(
        center,
        radii,
        id=None,
        description=None,
        linked_segmentation=None,
        tag_ids=None,
        **kwargs,
    ):
        return neuroglancer.ellipsoid_annotation(
            center=center,
            radii=radii,
            id=id,
            description=description,
            linked_segmentation=linked_segmentation,
            tag_ids=tag_ids,
        )

    @staticmethod
    def bounding_box_annotation(
        pointA,
        pointB,
        id=None,
        description=None,
        linked_segmentation=None,
        tag_ids=None,
        **kwargs,
    ):
        return neuroglancer.bounding_box_annotation(
            point_a=pointA,
            point_b=pointB,
            id=id,
            description=description,
            linked_segmentation=linked_segmentation,
            tag_ids=tag_ids,
        )

    @staticmethod
    def group_annotations(
        annotations,
        source=None,
        id=None,
        return_all=True,
        gather_linked_segmentations=True,
        share_linked_segmentations=False,
        children_visible=True,
        **kwargs,
    ):
        return neuroglancer.group_annotations(
            annotations,
            source=source,
            id=id,
            return_all=return_all,
            gather_linked_segmentations=gather_linked_segmentations,
            share_linked_segmentations=share_linked_segmentations,
            children_visible=children_visible,
        )



from abc import ABC, abstractmethod
from typing import Dict, List, Optional, Tuple, Union

from . import utils

SEGMENTATION_LAYER_TYPES = ["segmentation", "segmentation_with_graph"]


class EasyViewerBase(ABC):
    def __init__(self):
        self._default_viewer_url = None

    def __repr__(self):
        return self.as_url()

    def __repr_html__(self):
        return f'<a href="{self.as_url()}" target="_blank">Viewer</a>'

    @abstractmethod
    def load_url(self, url) -> None:
        pass

    def add_layers(
        self,
        image_layers: dict = {},
        segmentation_layers: dict = {},
        annotation_layers: dict = {},
        resolution: list = None,
    ) -> None:
        self.set_resolution(resolution)
        with self.txn() as s:
            for ln, kws in image_layers.items():
                s.layers[ln] = self._ImageLayer(**kws)
            for ln, kws in segmentation_layers.items():
                self._SegmentationLayer(s, **kws)
            for ln, kws in annotation_layers.items():
                s.layers[ln] = self._AnnotationLayer(**kws)

    @abstractmethod
    def _ImageLayer(self, source, **kwargs):
        pass

    @abstractmethod
    def _AnnotationLayer(self, **kwargs):
        pass

    @abstractmethod
    def _SegmentationLayer(self, source, **kwargs):
        pass

    def add_segmentation_layer(self, layer_name, source, **kwargs):
        """Add segmentation layer to viewer instance.

        Attributes:
            layer_name (str): name of layer to be displayed in neuroglancer ui.
            source (str): source of neuroglancer segment layer
        """
        with self.txn() as s:
            s.layers[layer_name] = self._SegmentationLayer(source=source, **kwargs)

    @abstractmethod
    def append_source_to_segmentation_layer(self, layer_name, source):
        pass

    @abstractmethod
    def add_skeleton_source(self, layer_name, source, shader_text=None):
        pass

    @abstractmethod
    def set_skeleton_shader(self, layer_name, shader_text=None):
        pass

    def add_image_layer(self, layer_name, source, contrast_range=None, **kwargs):
        """Add segmentation layer to viewer instance.

        Attributes:
            layer_name (str): name of layer to be displayed in neuroglancer ui.
            source (str): source of neuroglancer image layer
            contrast_range = (float, float): contrast range for image layer either in 0-1 (floats) or 0-255 (ints).
        """
        with self.txn() as s:
            s.layers[layer_name] = self._ImageLayer(source=source, **kwargs)
        if contrast_range is not None:
            self.add_contrast_shader(
                layer_name, black=contrast_range[0], white=contrast_range[1]
            )

    @abstractmethod
    def set_resolution(self, resolution) -> None:
        pass

    def add_contrast_shader(self, layer_name, black=0.0, white=1.0):
        shader_text = f"#uicontrol float black slider(min=0, max=1, default={black})\n#uicontrol float white slider(min=0, max=1, default={white})\nfloat rescale(float value) {{\n  return (value - black) / (white - black);\n}}\nvoid main() {{\n  float val = toNormalized(getDataValue());\n  if (val < black) {{\n    emitRGB(vec3(0,0,0));\n  }} else if (val > white) {{\n    emitRGB(vec3(1.0, 1.0, 1.0));\n  }} else {{\n    emitGrayscale(rescale(val));\n  }}\n}}\n"
        self.update_shader(layer_name, shader_text)

    def update_shader(self, layer_name, shader_text):
        with self.txn() as s:
            s.layers[layer_name]._json_data["shader"] = shader_text

    @abstractmethod
    def set_state_server(self, state_server) -> None:
        pass

    @abstractmethod
    def add_annotation_layer(
        self,
        layer_name=None,
        color=None,
        linked_segmentation_layer=None,
        filter_by_segmentation=False,
        brackets_show_segmentation=True,
        selection_shows_segmentation=True,
        tags=None,
    ) -> None:
        pass

    def set_annotation_layer_color(self, layer_name, color):
        """Set the color for the annotation layer"""
        if layer_name in [l.name for l in self.state.layers]:
            with self.txn() as s:
                s.layers[layer_name].annotationColor = utils.parse_color(color)
        else:
            pass

    def clear_annotation_layers(self, layer_names):
        with self.txn() as s:
            for ln in layer_names:
                s.layers[ln].annotations._data = []

    def add_annotations(
        self,
        layer_name: str,
        annotations: List,
    ):
        with self.txn() as s:
            s.layers[layer_name].annotations.extend(annotations)

    def add_multilayer_annotations(
        self,
        layer_anno_dict: Dict[str, List],
    ):
        """
        layer_anno_dict is a layer_name to annotation list dict.
        """
        with self.txn() as s:
            for ln, annos in layer_anno_dict.items():
                if annos is None:
                    continue
                s.layers[ln].annotations.extend(annos)

    def remove_annotations(self, layer_name, anno_ids):
        if isinstance(anno_ids, str):
            anno_ids = [anno_ids]
        try:
            with self.txn() as s:
                el = len(s.layers[layer_name].annotations)
                for anno in reversed(s.layers[layer_name].annotations):
                    el -= 1
                    if anno.id in anno_ids:
                        anno_ids.remove(anno.id)
                        s.layers[layer_name].annotations.pop(el)
                        if len(anno_ids) == 0:
                            break
        except:
            self.update_message("Could not remove annotation")

    @abstractmethod
    def add_annotation_tags(self, layer_name, tags):
        pass

    @property
    def url(self):
        return self.get_viewer_url()

    @abstractmethod
    def as_url(
        self,
        prefix: Optional[str] = None,
        as_html: Optional[bool] = False,
        link_text: Optional[str] = "Neuroglancer Link",
    ) -> str:
        pass

    @abstractmethod
    def set_selected_layer(self, layer_name):
        pass

    @abstractmethod
    def select_annotation(self, layer_name, anno_id):
        pass

    @property
    def layer_names(self):
        return [l.name for l in self.state.layers]

    def selected_objects(self, segmentation_layer):
        return list(self.state.layers[segmentation_layer].segments)

    @abstractmethod
    def add_selected_objects(
        self,
        segmentation_layer: str,
        oids: List[int],
        colors: Optional[Union[List, Dict]] = None,
    ) -> None:
        pass

    @abstractmethod
    def set_view_options(
        self,
        show_slices: Optional[bool] = None,
        layout: Optional[str] = None,
        show_axis_lines: Optional[bool] = None,
        show_scale_bar: Optional[bool] = None,
        orthographic: Optional[bool] = None,
        position: Optional[Tuple[float]] = None,
        zoom_image: Optional[float] = None,
        zoom_3d: Optional[float] = None,
        background_color: Optional[Tuple[float]] = None,
    ) -> None:
        pass

    @abstractmethod
    def set_segmentation_view_options(
        self,
        layer_name: str,
        alpha_selected: Optional[float] = None,
        alpha_3d: Optional[float] = None,
        alpha_unselected: Optional[float] = None,
        **kwargs,
    ):
        pass

    @abstractmethod
    def set_timestamp(
        self,
        layer_name,
        timestamp: Optional[int] = None,
    ):
        pass

    @abstractmethod
    def assign_colors(self, layer_name, seg_colors):
        """Assign colors to root ids in a segmentation layer

        Parameters
        ----------
        layer_name : str,
            Segmentation layer name
        seg_colors : dict
            dict with root ids as keys and colors as values.
        """

    @abstractmethod
    def set_multicut_points(
        self,
        layer_name,
        seg_id,
        points_red,
        points_blue,
        supervoxels_red=None,
        supervoxels_blue=None,
        focus=True,
    ):
        pass

    @staticmethod
    @abstractmethod
    def point_annotation(
        point,
        id=None,
        **kwargs,
    ):
        pass

    @staticmethod
    @abstractmethod
    def line_annotation(
        pointA,
        pointB,
        id=None,
        **kwargs,
    ):
        pass

    @staticmethod
    @abstractmethod
    def ellipsoid_annotation(
        center,
        radii,
        id=None,
        **kwargs,
    ):
        pass

    @classmethod
    def sphere_annotation(
        cls,
        center,
        radius,
        z_multiplier,
        **kwargs,
    ):
        return cls.ellipsoid_annotation(
            center,
            [radius, radius, radius * z_multiplier],
            **kwargs,
        )

    @staticmethod
    @abstractmethod
    def bounding_box_annotation(
        pointA,
        pointB,
        id=None,
        **kwargs,
    ):
        pass

    @staticmethod
    @abstractmethod
    def group_annotations(
        annotations,
        source=None,
        id=None,
        return_all=True,
        gather_linked_segmentations=True,
        share_linked_segmentations=False,
        children_visible=True,
    ):
        pass



import re
from collections.abc import Iterable

import numpy as np

from ..site_utils import (
    default_mainline_neuroglancer_base,
    default_seunglab_neuroglancer_base,
)

FALLBACK_SEUNGLAB_NGL_URL = default_seunglab_neuroglancer_base
FALLBACK_MAINLINE_NGL_URL = default_mainline_neuroglancer_base

SPLIT_SUFFIXES = ["x", "y", "z"]


def bucket_of_values(col, data, item_is_array=False, array_width=3):
    """
    Use to get a flat array of items when you don't know if it's already a collection or a collection of iterables.
    Parameters:
        dataseries: Pandas dataseries with either all items or all a collection of items.
                    If the item is expected to be an nd-array, use item_shape to define what an element is.
    """

    if len(data) == 0:
        return []

    dataseries = data[col]
    dataseries = dataseries[~dataseries.isnull()]

    if item_is_array:
        # If already an m x n array, just vstack. Else, need to stack every element first.
        if type(dataseries.iloc[0]) is np.ndarray:
            if len(data) > 1:
                return np.vstack(dataseries.values)
            else:
                return dataseries.values[0].reshape(1, -1)
        elif len(data) > 1:
            return np.vstack(dataseries.map(np.vstack)).reshape(-1, array_width)
        else:
            return np.vstack(dataseries).reshape(1, -1)
    elif isinstance(dataseries.iloc[0], Iterable):
        return np.concatenate(dataseries.values)
    else:
        return dataseries.values


def is_split_position(pt_col, df, suffixes=SPLIT_SUFFIXES):
    if pt_col in df.columns:
        return True
    prefix_found = []
    for suf in suffixes:
        prefix_found.append(
            np.any(
                [re.search(f"^{pt_col}_{suf}", col) is not None for col in df.columns]
            )
        )
    if np.all(prefix_found):
        return True
    else:
        raise ValueError(
            f'Point column "{pt_col}" not found directly or as split position'
        )


def is_split_split_position(pt_col, df, suffixes=SPLIT_SUFFIXES):
    is_split_split = []
    for suf in suffixes:
        split_name = pt_col.split("_")
        expected_name = "_".join(split_name[:-1]) + f"_{suf}_{split_name[-1]}"
        is_split_split.append(expected_name in df.columns)
    return np.all(is_split_split)


def assemble_split_points(pt_col, df, suffixes=SPLIT_SUFFIXES):
    cols = split_position_columns(pt_col, suffixes)
    return np.vstack(df[cols].values)


def split_position_columns(pt_col, suffixes=SPLIT_SUFFIXES):
    return [f"{pt_col}_{suf}" for suf in suffixes]



# @license
# Copyright 2017 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import base64
import collections
import numbers
import traceback

import numpy as np
import six

from . import viewer_state
from .json_wrappers import (
    JsonObjectWrapper,
    array_wrapper,
    optional,
    text_type,
    typed_list,
    typed_set,
    typed_string_map,
    wrapped_property,
)

_uint64_keys = frozenset(["t", "v"])
_map_entry_keys = frozenset(["key", "value"])

MapEntry = collections.namedtuple("MapEntry", ["key", "value"])


def layer_selected_value(x):
    if isinstance(x, numbers.Number):
        return x
    if isinstance(x, dict):
        if six.viewkeys(x) == _uint64_keys and x.get("t") == "u64":
            return int(x["v"])
        if six.viewkeys(x) == _map_entry_keys:
            return MapEntry(int(x["key"]), int(x["value"]))


LayerSelectedValues = typed_string_map(layer_selected_value)


class ScreenshotReply(JsonObjectWrapper):
    __slots__ = ()
    id = wrapped_property("id", text_type)
    image = wrapped_property("image", base64.b64decode)
    image_type = imageType = wrapped_property("imageType", text_type)


class ActionState(JsonObjectWrapper):
    __slots__ = ()
    viewer_state = viewerState = wrapped_property(
        "viewerState", viewer_state.ViewerState
    )
    selected_values = selectedValues = wrapped_property(
        "selectedValues", LayerSelectedValues
    )
    mouse_spatial_coordinates = mouseSpatialCoordinates = wrapped_property(
        "mouseSpatialCoordinates", optional(array_wrapper(np.float32, 3))
    )
    mouse_voxel_coordinates = mouseVoxelCoordinates = wrapped_property(
        "mouseVoxelCoordinates", optional(array_wrapper(np.float32, 3))
    )
    screenshot = wrapped_property("screenshot", optional(ScreenshotReply))


class Actions:
    def __init__(self, set_config):
        self._action_handlers = dict()
        self._set_config = set_config

    def add(self, name, handler):
        self._action_handlers.setdefault(name, set()).add(handler)
        self._update_config()

    def clear(self):
        self._action_handlers.clear()
        self._update_config()

    def remove(self, name, handler):
        handlers = self._action_handlers.get(name)
        if handlers is None:
            return
        handlers.remove(handler)
        if not handlers:
            del self._action_handlers[name]
        self._update_config()

    def _update_config(self):
        self._set_config(six.viewkeys(self._action_handlers))

    def invoke(self, name, state):
        state = ActionState(state)
        handlers = self._action_handlers.get(name)
        if handlers is not None:
            for handler in handlers:
                try:
                    handler(state)
                except:
                    traceback.print_exc()


EventActionMap = typed_string_map(text_type)


class InputEventBindings(JsonObjectWrapper):
    __slots__ = ()
    viewer = wrapped_property("viewer", EventActionMap)
    slice_view = sliceView = wrapped_property("sliceView", EventActionMap)
    perspective_view = perspectiveView = wrapped_property(
        "perspectiveView", EventActionMap
    )
    data_view = dataView = wrapped_property("dataView", EventActionMap)


class PrefetchState(JsonObjectWrapper):
    __slots__ = ()
    supports_validation = True
    priority = wrapped_property("priority", optional(int, 0))
    state = wrapped_property("state", viewer_state.ViewerState)


class ScaleBarOptions(JsonObjectWrapper):
    __slots__ = ()
    supports_validation = True
    scale_factor = scaleFactor = wrapped_property("scaleFactor", optional(float, 1))
    text_height_in_pixels = textHeightInPixels = wrapped_property(
        "textHeightInPixels", optional(float, 15)
    )
    bar_height_in_pixels = barHeightInPixels = wrapped_property(
        "barHeightInPixels", optional(float, 8)
    )
    bar_top_margin_in_pixels = barTopMarginInPixels = wrapped_property(
        "barTopMarginInPixels", optional(float, 5)
    )
    font_name = fontName = wrapped_property(
        "fontName", optional(text_type, "sans-serif")
    )
    padding_in_pixels = paddingInPixels = wrapped_property(
        "paddingInPixels", optional(float, 2)
    )
    max_width_in_pixels = maxWidthInPixels = wrapped_property(
        "maxWidthInPixels", optional(int, 100)
    )
    max_width_fraction = maxWidthFraction = wrapped_property(
        "maxWidthFraction", optional(float, 0.25)
    )
    left_pixel_offset = leftPixelOffset = wrapped_property(
        "leftPixelOffset", optional(int, 10)
    )
    bottom_pixel_offset = bottomPixelOffset = wrapped_property(
        "bottomPixelOffset", optional(int, 10)
    )


class ConfigState(JsonObjectWrapper):
    __slots__ = ()
    credentials = wrapped_property("credentials", typed_string_map(dict))
    actions = wrapped_property("actions", typed_set(text_type))
    input_event_bindings = inputEventBindings = wrapped_property(
        "inputEventBindings", InputEventBindings
    )
    status_messages = statusMessages = wrapped_property(
        "statusMessages", typed_string_map(text_type)
    )
    source_generations = sourceGenerations = wrapped_property(
        "sourceGenerations", typed_string_map(int)
    )
    screenshot = wrapped_property("screenshot", optional(text_type))
    show_ui_controls = showUIControls = wrapped_property(
        "showUIControls", optional(bool, True)
    )
    show_location = showLocation = wrapped_property(
        "showLocation", optional(bool, True)
    )
    show_layer_panel = showLayerPanel = wrapped_property(
        "showLayerPanel", optional(bool, True)
    )
    show_help_button = showHelpButton = wrapped_property(
        "showHelpButton", optional(bool, True)
    )
    show_panel_borders = showPanelBorders = wrapped_property(
        "showPanelBorders", optional(bool, True)
    )
    scale_bar_options = scaleBarOptions = wrapped_property(
        "scaleBarOptions", ScaleBarOptions
    )
    viewer_size = viewerSize = wrapped_property(
        "viewerSize", optional(array_wrapper(np.int64, 2))
    )
    prefetch = wrapped_property("prefetch", typed_list(PrefetchState))


class PrivateState(JsonObjectWrapper):
    __slots__ = ()
    credentials = wrapped_property("credentials", typed_string_map(optional(int)))



import re
from typing import Optional, Tuple, Union

import numpy as np
import pandas as pd
from box import Box

from ..easyviewer.ev_base.base import SEGMENTATION_LAYER_TYPES


def _is_spelunker_state(state) -> bool:
    """Check if a state is a spelunker state or not."""
    return "dimensions" in state.keys()


def _data_resolution(state, layer=None) -> Tuple[float, float, float]:
    """Get the data resolution from the state"""
    if _is_spelunker_state(state):
        if layer is None:
            raise ValueError("Layer must be specified for spelunker state")
        l = get_layer(state, layer)
        dims = l["source"]["transform"]["outputDimensions"]
        dim_x = dims["x"][0] * 10**9
        dim_y = dims["y"][0] * 10**9
        dim_z = dims["z"][0] * 10**9
        return (dim_x, dim_y, dim_z)
    else:
        dims = state["navigation"]["pose"]["position"]["voxelSize"]
    return tuple(dims)


def layer_names(state: dict, include_archived: bool = True) -> list:
    """Get all layer names in the state

    Parameters
    ----------
    state : dict
        Neuroglancer state as a JSON dict
    include_archived : bool, optional
        If True, include archived layers. By default True

    Returns
    -------
    names : list
        List of layer names
    """
    if include_archived:
        return [l["name"] for l in state["layers"]]
    else:
        return [l["name"] for l in state["layers"] if not l.get("archived", False)]


def image_layers(state: dict, include_archived: bool = True) -> list:
    """Get all image layer names in the state

    Parameters
    ----------
    state : dict
        Neuroglancer state as a JSON dict
    include_archived : bool, optional
        If True, include archived layers. By default True

    Returns
    -------
    names : list
        List of layer names
    """
    if include_archived:
        return [l["name"] for l in state["layers"] if l["type"] == "image"]
    else:
        return [
            l["name"]
            for l in state["layers"]
            if l["type"] == "image"
            if not l.get("archived", False)
        ]


def segmentation_layers(state: dict, include_archived: bool = True) -> list:
    """Get all segmentation layer names in the state

    Parameters
    ----------
    state : dict
        Neuroglancer state as a JSON dict

    Returns
    -------
    names : list
        List of layer names
    """
    if include_archived:
        return [
            l["name"] for l in state["layers"] if l["type"] in SEGMENTATION_LAYER_TYPES
        ]
    else:
        return [
            l["name"]
            for l in state["layers"]
            if l["type"] in SEGMENTATION_LAYER_TYPES
            if not l.get("archived", False)
        ]


def annotation_layers(state: dict, include_archived: bool = True) -> list:
    """Get all annotation layer names in the state

    Parameters
    ----------
    state : dict
        Neuroglancer state as a JSON dict
    include_archived : bool, optional
        If True, include archived layers. By default True

    Returns
    -------
    names : list
        List of layer names
    """
    if include_archived:
        return [l["name"] for l in state["layers"] if l["type"] == "annotation"]
    else:
        return [
            l["name"]
            for l in state["layers"]
            if l["type"] == "annotation"
            if not l.get("archived", False)
        ]


def layer_source(state: dict, layer_name: str) -> Union[str, list]:
    """Get url source or list of sources for a layer

    Parameters
    ----------
    state : dict
        Neuroglancer state as a JSON dict
    layer_name : str
        Name of layer

    Returns
    -------
    str
        URL source or list of URL sources
    """
    l = get_layer(state, layer_name)
    source = l.get("source", None)
    if isinstance(source, list):
        source = [s.get("url", None) for s in source]
    elif isinstance(source, dict):
        source = source.get("url", None)
    return source


def tag_dictionary(state: dict, layer_name: str) -> dict:
    """Get the tag id to string dictionary for a layer

    Parameters
    ----------
    state : dict

    layer_name : [type]
        [description]

    Returns
    -------
    [type]
        [description]
    """
    l = get_layer(state, layer_name)
    if _is_spelunker_state(state):
        tag_props = l.get("annotationProperties", [])
        tags = {}
        for ii, tp in enumerate(tag_props):
            if (tag_str := tp.get("tag")) is not None:
                m = re.match(r"^tag(\d+)$", tp.get("id", ""))
                if m:
                    tags[int(m.groups()[0])] = tag_str
    else:
        taginfo = l.get("annotationTags", [])
        tags = {}
        for t in taginfo:
            tags[t["id"]] = t["label"]
    return tags


def get_layer(state: dict, layer_name: str) -> dict:
    """Gets the contents of the layer based on the layer name.

    Parameters
    ----------
    state : dict
        Neuroglancer state as a JSON dict
    layer_name : str
        Name of layer

    Returns
    -------
    layer : dict
        Layer data contents
    """
    layer_ind = layer_names(state).index(layer_name)
    return state["layers"][layer_ind]


def view_settings(state: dict) -> dict:
    """Get all data about the view state in neuroglancer: position,
    image zoom, orientation and zoom of the 3d view, and voxel size.

    Parameters
    ----------
    state : dict
        Neuroglancer state as JSON dict

    Returns
    -------
    view : dict
        Dictionary with keys: position, zoomFactor, perspectiveOrientation,
        perspectiveZoom, and voxelSize
    """
    view = {}
    view["position"] = state["navigation"]["pose"]["position"]["voxelCoordinates"]
    view["zoomFactor"] = state["navigation"].get("zoomFactor", None)
    view["perspectiveOrientation"] = state.get("perspectiveOrientation")
    view["perspectiveZoom"] = state.get("perspectiveZoom")
    view["voxelSize"] = state["navigation"]["pose"]["position"]["voxelSize"]
    return view


def get_selected_ids(
    state: dict, layer: Optional[str] = None, return_nonvisible: bool = False
) -> Union[list, Tuple[list, list]]:
    """Get a list of selected ids in a segmentation layer

    Parameters
    ----------
    state : dict
        State dict
    layer : str, optional
        Segmentation layer name, if needed. If None and only one segmentation layer
        is present, default to it. By default None
    return_nonvisible : bool, optional
        If True, also return non-visible ids. By default False

    Returns
    -------
    list
        List of root ids.
    """
    seg_layers = segmentation_layers(state)
    if layer is None:
        if len(seg_layers) == 1:
            layer = seg_layers[0]
        else:
            raise ValueError(
                "Segmentation layer must be specified since the state has more than one"
            )
    l = get_layer(state, layer)
    selected_ids = [int(s) for s in l["segments"] if s[0] != "!"]
    invisible_ids = [int(s[1:]) for s in l["segments"] if s[0] == "!"]
    if return_nonvisible:
        return selected_ids, invisible_ids
    else:
        return selected_ids


def _get_type_annotations(state, layer_name, type):
    l = get_layer(state, layer_name)
    annos = l.get("annotations", [])
    return [anno for anno in annos if anno["type"] == type]


def _get_point_annotations(state, layer_name):
    return _get_type_annotations(state, layer_name, "point")


def _get_sphere_annotations(state, layer_name):
    return _get_type_annotations(state, layer_name, "ellipsoid")


def _get_line_annotations(state, layer_name):
    return _get_type_annotations(state, layer_name, "line")


def _get_bbox_annotations(state, layer_name):
    return _get_type_annotations(state, layer_name, "axis_aligned_bounding_box")


def _get_group_annotations(state, layer_name):
    return _get_type_annotations(state, layer_name, "collection")


def _extract_point_data(annos):
    return [[anno.get("point") for anno in annos]]


def _extract_sphere_data(annos):
    pt = [anno.get("center") for anno in annos]
    rad = [anno.get("radii") for anno in annos]
    return [pt, rad]


def _extract_line_data(annos):
    ptA = [anno.get("pointA") for anno in annos]
    ptB = [anno.get("pointB") for anno in annos]
    return [ptA, ptB]


def _extract_bbox_data(annos):
    ptA = [anno.get("pointA") for anno in annos]
    ptB = [anno.get("pointB") for anno in annos]
    return [ptA, ptB]


def _extract_group_data(annos):
    pt = [anno.get("source") for anno in annos]
    anno_id = [anno.get("id") for anno in annos]
    return [pt, anno_id]


_get_map = {
    "point": _get_point_annotations,
    "line": _get_line_annotations,
    "sphere": _get_sphere_annotations,
    "bbox": _get_bbox_annotations,
    "group": _get_group_annotations,
}

_extraction_map = {
    "point": _extract_point_data,
    "line": _extract_line_data,
    "sphere": _extract_sphere_data,
    "bbox": _extract_bbox_data,
    "group": _extract_group_data,
}


def _flatten_list_of_strings(l):
    out = []
    for x in l:
        if isinstance(x, list):
            out.extend(x)
        else:
            out.append(x)
    return out


def _extract_segments(anno):
    seg_data = anno.get("segments", [])
    seg_list = _flatten_list_of_strings(seg_data)
    return [int(x) for x in seg_list]


def _generic_annotations(
    state, layer_name, description, linked_segmentations, tags, group, anno_type
):
    annos = _get_map[anno_type](state, layer_name)
    out = _extraction_map[anno_type](annos)
    if description:
        desc = [anno.get("description", None) for anno in annos]
        out.append(desc)
    if linked_segmentations:
        out.append([_extract_segments(anno) for anno in annos])
    if tags:
        if not _is_spelunker_state(state):
            tag_list = [anno.get("tagIds", []) for anno in annos]
        else:
            tag_list = [
                [ii for ii, val in enumerate(anno["props"]) if val == 1]
                for anno in annos
            ]
        out.append(tag_list)
    if group:
        group_ids = [anno.get("parentId") for anno in annos]
        out.append(group_ids)
    if len(out) == 1:
        return out[0]
    else:
        return out


def point_annotations(
    state: dict,
    layer_name: str,
    description: bool = False,
    linked_segmentations: bool = False,
    tags: bool = False,
    group: bool = False,
):
    """Get all point annotation points and other info from a layer.

    Parameters
    ----------
    state : dict
        Neuroglancer state as JSON dict
    layer_name : str
        Layer name
    description : bool, optional
        If True, also returns descriptions as well. By default False
    linked_segmentations : bool, optional
        If True, also returns list of linked segmentations, by default False
    tags : bool, optional
        If True, also returns list of tags, by default False

    Returns
    -------
    anno_points : list
        List of N 3-element points (as list)
    anno_descriptions : list
        List of N strings (or None), only returned if description=True.
    anno_linked_segmentations : list
        List of N lists of object ids. Only returned if linked_segmentations=True.
    anno_tags : list
        List of N lists of tag ids. Only returned if tags=True.
    anno_group : list
        List of group ids (as string) or None for annotations. Only returned if group=True
    """
    return _generic_annotations(
        state,
        layer_name,
        description=description,
        linked_segmentations=linked_segmentations,
        tags=tags,
        group=group,
        anno_type="point",
    )


def line_annotations(
    state: dict,
    layer_name: str,
    description: bool = False,
    linked_segmentations: bool = False,
    tags: bool = False,
    group: bool = False,
):
    """Get all line annotation points and other info from a layer.

    Parameters
    ----------
    state : dict
        Neuroglancer state as JSON dict
    layer_name : str
        Layer name
    description : bool, optional
        If True, also returns descriptions as well. By default False
    linked_segmentations : bool, optional
        If True, also returns list of linked segmentations, by default False
    tags : bool, optional
        If True, also returns list of tags, by default False

    Returns
    -------
    anno_points_A : list
        List of N 3-element points (as list) of the first point in each line.
    anno_points_B : list
        List of N 3-element points (as list) of the second point in each line.
    anno_descriptions : list
        List of N strings (or None), only returned if description=True.
    anno_linked_segmentations : list
        List of N lists of object ids. Only returned if linked_segmentations=True.
    anno_tags : list
        List of N lists of tag ids. Only returned if tags=True.
    anno_group : list
        List of group ids (as string) or None for annotations. Only returned if group=True
    """
    return _generic_annotations(
        state, layer_name, description, linked_segmentations, tags, group, "line"
    )


def bbox_annotations(
    state: dict,
    layer_name: str,
    description: bool = False,
    linked_segmentations: bool = False,
    tags: bool = False,
    group: bool = False,
):
    """Get all bounding box annotation points and other info from a layer.

    Parameters
    ----------
    state : dict
        Neuroglancer state as JSON dict
    layer_name : str
        Layer name
    description : bool, optional
        If True, also returns descriptions as well. By default False
    linked_segmentations : bool, optional
        If True, also returns list of linked segmentations, by default False
    tags : bool, optional
        If True, also returns list of tags, by default False

    Returns
    -------
    anno_points_A : list
        List of N 3-element points (as list) of the first point in each bbox.
    anno_points_B : list
        List of N 3-element points (as list) of the second point in each bbox.
    anno_descriptions : list
        List of N strings (or None), only returned if description=True.
    anno_linked_segmentations : list
        List of N lists of object ids. Only returned if linked_segmentations=True.
    anno_tags : list
        List of N lists of tag ids. Only returned if tags=True.
    anno_group : list
        List of group ids (as string) or None for annotations. Only returned if group=True
    """
    return _generic_annotations(
        state, layer_name, description, linked_segmentations, tags, group, "bbox"
    )


def sphere_annotations(
    state: dict,
    layer_name: str,
    description: bool = False,
    linked_segmentations: bool = False,
    tags: bool = False,
    group: bool = False,
):
    """Get all sphere annotation points and other info from a layer.

    Parameters
    ----------
    state : dict
        Neuroglancer state as JSON dict
    layer_name : str
        Layer name
    description : bool, optional
        If True, also returns descriptions as well. By default False
    linked_segmentations : bool, optional
        If True, also returns list of linked segmentations, by default False
    tags : bool, optional
        If True, also returns list of tags, by default False

    Returns
    -------
    anno_points : list
        List of N 3-element center points (as list)
    radius_points : list
        List of N 3-element radii for each axis of the ellipsoid.
    anno_descriptions : list
        List of N strings (or None), only returned if description=True.
    anno_linked_segmentations : list
        List of N lists of object ids. Only returned if linked_segmentations=True.
    anno_tags : list
        List of N lists of tag ids. Only returned if tags=True.
    anno_group : list
        List of group ids (as string) or None for annotations. Only returned if group=True
    """
    return _generic_annotations(
        state, layer_name, description, linked_segmentations, tags, group, "sphere"
    )


def group_annotations(
    state: dict,
    layer_name: str,
    description: bool = False,
    linked_segmentations: bool = False,
    tags: bool = False,
):
    """All group annotations and their associated points

    Parameters
    ----------
    state : dict
        Neuroglancer state as JSON dict
    layer_name : str
        Annotation layer name
    description : bool, optional
        If True, also returns descriptions as well. By default False
    linked_segmentations : bool, optional
        If True, also returns list of linked segmentations, by default False
    tags : bool, optional
        If True, also returns list of tags, by default False

    Returns
    -------
    group_points : list
        List of N 3-element points
    group_id: list
        List of N id strings for groups.
    anno_descriptions : list
        List of N strings (or None), only returned if description=True.
    anno_linked_segmentations : list
        List of N lists of object ids. Only returned if linked_segmentations=True.
    anno_tags : list
        List of N lists of tag ids. Only returned if tags=True.
    """
    return _generic_annotations(
        state,
        layer_name,
        description=description,
        linked_segmentations=linked_segmentations,
        tags=tags,
        group=False,
        anno_type="group",
    )


def extract_multicut(
    state: dict, seg_layer: str = None
) -> Tuple[np.array, np.array, np.array, int]:
    """Extract information entered into the multicut graph operation

    Parameters
    ----------
    state : dict
        Neuroglancer state
    seg_layer : str, optional
        Name of a segmentation layer or None. If None, the function will check how many segmentation
        layers there are and, if only one exits, choose it. If more than one segmentation layer is present,
        it errors. By default None

    Returns
    -------
    pts: np.array
        Nx3 array of points selected
    side: np.array
        N array with 'source' or 'sink', depending on which side the point is on.
    svids: np.array
        N array with selected supervoxel. If only points are selected (e.g. via clicking on the mesh),
        the value will be NaN.
    root_id: int
        Root id of the object to split
    """
    if seg_layer is None:
        seg_layers = segmentation_layers(state)
        if len(seg_layers) == 1:
            seg_layer = seg_layers[0]
        else:
            raise ValueError(
                "State has multiple segmentation layers. Please specify the layer to use."
            )
    l = get_layer(state, "seg")
    pts = []
    svids = []
    side = []
    root_id = None

    source_data = l["graphOperationMarker"][0]
    for anno in source_data["annotations"]:
        pts.append(anno["point"])
        side.append("source")
        if root_id is None:
            root_id = int(anno["segments"][1])
        svid = int(anno["segments"][0])
        if svid != root_id:
            svids.append(svid)
        else:
            # This means that no supervoxel was directly selected
            svids.append(np.nan)

    sink_data = l["graphOperationMarker"][1]
    for anno in sink_data["annotations"]:
        pts.append(anno["point"])
        side.append("sink")
        svid = int(anno["segments"][0])
        if root_id is None:
            root_id = int(anno["segments"][1])
        if svid != root_id:
            svids.append(svid)
        else:
            # This means that no supervoxel was directly selected
            svids.append(np.nan)

    return np.array(pts), np.array(side), np.array(svids), root_id


def _concat_list(d):
    d_out = []
    for x in d:
        for y in x:
            d_out.append(y)
    return d_out


def _parse_layer_dataframe(
    state, ln, expand_tags, point_resolution=None, split_points=False
):
    lns = []
    points = []
    anno_types = []
    pointBs = []
    linked_segs = []
    tags = []
    group_ids = []
    descs = []
    if point_resolution:
        data_resolution = np.array(_data_resolution(state, layer=ln))
        scaling = np.array(point_resolution) / data_resolution
    else:
        scaling = np.array([1, 1, 1])
    p_pt, p_desc, p_seg, p_tag, p_grp = point_annotations(
        state,
        ln,
        description=True,
        linked_segmentations=True,
        tags=True,
        group=True,
    )
    p_pt = (np.array(p_pt).reshape(-1, 3) * scaling).tolist()
    n_p_pts = len(p_pt)

    p_type = ["point"] * n_p_pts
    p_ln = [ln] * n_p_pts
    p_ptB = [np.nan] * n_p_pts

    lns.append(p_ln)
    points.append(p_pt)
    anno_types.append(p_type)
    pointBs.append(p_ptB)
    linked_segs.append(p_seg)
    tags.append(p_tag)
    group_ids.append(p_grp)
    descs.append(p_desc)

    # Lines
    l_ptA, l_ptB, l_desc, l_seg, l_tag, l_grp = line_annotations(
        state,
        ln,
        description=True,
        linked_segmentations=True,
        tags=True,
        group=True,
    )
    l_ptA = (np.array(l_ptA).reshape(-1, 3) * scaling).tolist()
    l_ptB = (np.array(l_ptB).reshape(-1, 3) * scaling).tolist()
    n_l_pts = len(l_ptA)
    l_type = ["line"] * n_l_pts
    l_ln = [ln] * n_l_pts

    lns.append(l_ln)
    points.append(l_ptA)
    anno_types.append(l_type)
    pointBs.append(l_ptB)
    linked_segs.append(l_seg)
    tags.append(l_tag)
    group_ids.append(l_grp)
    descs.append(l_desc)

    # Spheres
    s_ptA, s_ptB, s_desc, s_seg, s_tag, s_grp = sphere_annotations(
        state,
        ln,
        description=True,
        linked_segmentations=True,
        tags=True,
        group=True,
    )
    s_ptA = (np.array(s_ptA).reshape(-1, 3) * scaling).tolist()
    s_ptB = (np.array(s_ptB).reshape(-1, 3) * scaling).tolist()
    n_s_pts = len(s_ptA)
    s_type = ["sphere"] * n_s_pts
    s_ln = [ln] * n_s_pts

    lns.append(s_ln)
    points.append(s_ptA)
    anno_types.append(s_type)
    pointBs.append(s_ptB)
    linked_segs.append(s_seg)
    tags.append(s_tag)
    group_ids.append(s_grp)
    descs.append(s_desc)

    # Bboxes
    b_ptA, b_ptB, b_desc, b_seg, b_tag, b_grp = bbox_annotations(
        state,
        ln,
        description=True,
        linked_segmentations=True,
        tags=True,
        group=True,
    )
    b_ptA = (np.array(b_ptA).reshape(-1, 3) * scaling).tolist()
    b_ptB = (np.array(b_ptB).reshape(-1, 3) * scaling).tolist()
    n_b_pts = len(b_ptA)
    b_type = ["bbox"] * n_b_pts
    b_ln = [ln] * n_b_pts

    lns.append(b_ln)
    points.append(b_ptA)
    anno_types.append(b_type)
    pointBs.append(b_ptB)
    linked_segs.append(b_seg)
    tags.append(b_tag)
    group_ids.append(b_grp)
    descs.append(b_desc)

    df = pd.DataFrame(
        {
            "layer": _concat_list(lns),
            "anno_type": _concat_list(anno_types),
            "point": _concat_list(points),
            "pointB": _concat_list(pointBs),
            "linked_segmentation": _concat_list(linked_segs),
            "tags": _concat_list(tags),
            "group_id": _concat_list(group_ids),
            "description": _concat_list(descs),
        }
    )
    if split_points:
        df["point_x"] = df["point"].apply(lambda x: x[0])
        df["point_y"] = df["point"].apply(lambda x: x[1])
        df["point_z"] = df["point"].apply(lambda x: x[2])
        df["pointB_x"] = df["pointB"].apply(
            lambda x: x[0] if not np.isnan(x) else np.nan
        )
        df["pointB_y"] = df["pointB"].apply(
            lambda x: x[1] if not np.isnan(x) else np.nan
        )
        df["pointB_z"] = df["pointB"].apply(
            lambda x: x[2] if not np.isnan(x) else np.nan
        )
        df.drop(columns=["point", "pointB"], inplace=True)

    if expand_tags:
        tag_dict = tag_dictionary(state, ln)
        for k, v in tag_dict.items():
            df[v] = df["tags"].apply(lambda x: k in x)
    return df


def selection_dataframe(state: dict, include_archived: bool = True) -> pd.DataFrame:
    """Return a dataframe with a row for each selected id across segmentation layers.

    Parameters
    ----------
    state : dict
        Neuroglancer state
    include_archived: bool, optional
        If True, include archived layers. By default True

    Returns
    -------
    pd.DataFrame
        Dataframe with columns: layer, id, visible
        Visible only applies in Spelunker states.
    """
    layer_names = segmentation_layers(state, include_archived=include_archived)
    ls = []
    selected_ids = []
    is_visible = []
    for l in layer_names:
        ids_vis, ids_nonvis = get_selected_ids(state, layer=l, return_nonvisible=True)
        ls.extend([l] * len(ids_vis + ids_nonvis))
        selected_ids.extend(ids_vis + ids_nonvis)
        is_visible.extend([True] * len(ids_vis) + [False] * len(ids_nonvis))
    return pd.DataFrame({"layer": ls, "id": selected_ids, "visible": is_visible})


def layer_dataframe(state: dict, include_archived: bool = True) -> pd.DataFrame:
    """Return a dataframe with a row for each layer in the state.

    Parameters
    ----------
    state : dict
        Neuroglancer state
    include_archived : bool, optional
        If True, include archived layers. By default True

    Returns
    -------
    pd.DataFrame
        Dataframe with columns layer, type, source
    """
    names = layer_names(state, include_archived=include_archived)
    types = [get_layer(state, l)["type"] for l in names]
    source = [layer_source(state, l) for l in names]
    archived = [get_layer(state, l).get("archived", False) for l in names]
    return pd.DataFrame(
        {"layer": names, "type": types, "source": source, "archived": archived}
    )


def annotation_dataframe(
    state: dict,
    expand_tags: bool = False,
    point_resolution: Optional[Tuple[float, float, float]] = None,
    split_points: bool = False,
    include_archived: bool = True,
) -> pd.DataFrame:
    """Return a dataframe with all annotations across all annotation layers in the state.

    Parameters
    ----------
    state : dict
        Neuroglancer state dictionary
    expand_tags : bool, optional
        If True, expand tags into separate boolean columns named by the tag label. By default False.
        Note that if tag labels are duplicated in multiple layers, the values will appear in the same column.
    point_resolution : Tuple[float, float, float], optional
        If provided, points will be rescaled to the provided x, y, z resolution (in nanometers). By default None.
    split_points : bool, optional
        If True, points will be split into separate x, y, z columns. By default False.
    include_archived : bool, optional
        If True, include archived layers. By default True

    Returns
    -------
    pd.DataFrame
        Dataframe with columns layer, anno_type, point, pointB, linked_segmentation, tags, anno_id, group_id, description.
        If expand_tags is True, an additional column will be added for each tag.
    """
    dfs = [
        _parse_layer_dataframe(
            state,
            ln,
            expand_tags,
            point_resolution=point_resolution,
            split_points=split_points,
        )
        for ln in annotation_layers(state, include_archived=include_archived)
    ]
    return pd.concat(dfs, ignore_index=True)


class StateParser:
    def __init__(self, state: dict):
        """Convenience class for parsing neuroglancer states.

        Parameters
        ----------
        state : dict
            Neuroglancer state as a JSON dict
        """
        self._state = Box(state)

    @property
    def state(self):
        return self._state

    @property
    def layer_names(self) -> list:
        return layer_names(self.state)

    @property
    def is_spelunker_state(self):
        return _is_spelunker_state(self.state)

    @property
    def image_layers(self) -> list:
        return image_layers(self.state)

    @property
    def segmentation_layers(self) -> list:
        return segmentation_layers(self.state)

    @property
    def annotation_layers(self) -> list:
        return annotation_layers(self.state)

    def annotation_dataframe(
        self,
        expand_tags: bool = False,
        point_resolution=None,
        split_points: bool = False,
        include_archived: bool = True,
    ) -> pd.DataFrame:
        """Get a dataframe with all annotations across all annotation layers in the state.

        Parameters
        ----------
        expand_tags : bool, optional
            If True, expand tags into separate boolean columns named by the tag label. By default False.
            Note that if tag labels are duplicated in multiple layers, the values will appear in the same column.
        point_resolution : Tuple[float, float, float], optional
            If provided, points will be rescaled to the provided x, y, z resolution (in nanometers). By default None.
        split_points : bool, optional
            If True, points will be split into separate x, y, z columns. By default False.
        include_archived : bool, optional
            If True, include archived layers. By default True

        Returns
        -------
        pd.DataFrame
            Dataframe with columns layer, anno_type, point, pointB, linked_segmentation, tags, anno_id, group_id, description.
            If expand_tags is True, an additional column will be added for each tag.
        """

        return annotation_dataframe(
            self.state,
            expand_tags=expand_tags,
            point_resolution=point_resolution,
            split_points=split_points,
            include_archived=True,
        )

    def layer_dataframe(self, include_archived: bool = True) -> pd.DataFrame:
        """Get a dataframe with a row for each layer in the state.

        Parameters
        ----------
        include_archived : bool, optional
            If True, include archived layers. By default True

        Returns
        -------
        pd.DataFrame
            Dataframe with columns layer, type, source containing layer information.
        """
        return layer_dataframe(self.state, include_archived=include_archived)

    def selection_dataframe(self, include_archived: bool = True) -> pd.DataFrame:
        """Get a dataframe with a row for each selected id across segmentation layers.
        Parameters
        ----------
        include_archived: bool, optional
            If True, include archived layers. By default True

        Returns
        -------
        pd.DataFrame
            Dataframe with columns: layer, id, visible
            Visible only applies in Spelunker states.
        """
        return selection_dataframe(self.state, include_archived=include_archived)



from neuroglancer import (
    AxisAlignedBoundingBoxAnnotation,
    EllipsoidAnnotation,
    LineAnnotation,
    PointAnnotation,
)


def convert_annotation(anno):
    if anno.type == "point":
        out_anno = convert_point_annotation(anno)
    elif anno.type == "line":
        out_anno = convert_line_annotation(anno)
    elif anno.type == "axis_aligned_bounding_box":
        out_anno = convert_bbox_annotation(anno)
    elif anno.type == "ellipsoid":
        out_anno = convert_sphere_annotation(anno)
    else:
        out_anno = None
    return out_anno


def convert_point_annotation(anno):
    out_anno = PointAnnotation(
        point=anno.point.tolist(),
        id=anno.id,
        description=anno.description,
    )
    if len(anno.segments) > 0:
        out_anno.segments = [[int(x) for x in anno.segments]]
    return out_anno


def convert_line_annotation(anno):
    out_anno = LineAnnotation(
        point_a=anno.point_a.tolist(),
        point_b=anno.point_b.tolist(),
        id=anno.id,
        description=anno.description,
    )
    if len(anno.segments) > 0:
        out_anno.segments = [[int(x) for x in anno.segments]]
    return out_anno


def convert_bbox_annotation(anno):
    # AxisAlignedBoundingBox happens to use the same parameters as LineAnnotation
    out_anno = AxisAlignedBoundingBoxAnnotation(
        point_a=anno.point_a.tolist(),
        point_b=anno.point_b.tolist(),
        id=anno.id,
        description=anno.description,
    )
    if len(anno.segments) > 0:
        out_anno.segments = [[int(x) for x in anno.segments]]
    return out_anno


def convert_sphere_annotation(anno):
    out_anno = EllipsoidAnnotation(
        center=anno.center.tolist(),
        radii=anno.radii.tolist(),
        id=anno.id,
        description=anno.description,
    )
    if len(anno.segments) > 0:
        out_anno.segments = [[int(x) for x in anno.segments]]
    return out_anno



import numbers
import re
from urllib.parse import urlparse

import numpy as np
import pandas as pd
import webcolors

from ...site_utils import (
    is_mainline,
    is_seunglab,
)


def omit_nones(seg_list):
    if seg_list is None or np.all(pd.isna(seg_list)):
        return []
    seg_list = np.atleast_1d(seg_list)
    seg_list = list(filter(lambda x: x is not None, seg_list))
    if len(seg_list) == 0:
        return []
    else:
        return seg_list


def parse_color(clr):
    if clr is None:
        return None

    if isinstance(clr, numbers.Number):
        clr = (clr, clr, clr)

    if isinstance(clr, str):
        hex_match = r"\#[0123456789abcdef]{6}"
        if re.match(hex_match, clr.lower()):
            return clr
        else:
            return webcolors.name_to_hex(clr)
    else:
        return webcolors.rgb_to_hex([int(255 * x) for x in clr])


def parse_graphene_header(source, target):
    qry = urlparse(source)
    if qry.scheme == "graphene":
        if is_seunglab(target):
            return _parse_to_seunglab(qry)
        elif is_mainline(target):
            return _parse_to_mainline(qry)
    else:
        return source


def parse_graphene_image_url(source, target):
    qry = urlparse(source)
    if qry.scheme == "graphene":
        if is_mainline(target):
            return _parse_to_mainline_imagery(qry)
        elif is_seunglab(target):
            return _parse_to_seunglab(qry)
    else:
        return source


def _parse_to_seunglab(qry):
    return f"{qry.scheme}://https:{qry.path}"


def _parse_to_mainline(qry):
    if "https" in qry.netloc:
        return f"{qry.scheme}://middleauth+https:{qry.path}"
    else:
        return f"{qry.scheme}://middleauth+http:{qry.path}"


def _parse_to_mainline_imagery(qry):
    if qry.scheme == "graphene":
        if "https" in qry.netloc:
            return f"precomputed://middleauth+https:{qry.path}"
        else:
            return f"precomputed://middleauth+http:{qry.path}"
    else:
        return qry.geturl()



from __future__ import annotations

from collections.abc import Iterable
from typing import TYPE_CHECKING, Literal, Optional, Union
from warnings import warn

import pandas as pd
from caveclient import CAVEclient
from caveclient.endpoints import fallback_ngl_endpoint
from IPython.display import HTML

from ..site_utils import neuroglancer_url
from .layers import (
    AnnotationLayerConfig,
    ImageLayerConfig,
    SegmentationLayerConfig,
)
from .mappers import LineMapper, PointMapper
from .statebuilder import ChainedStateBuilder, StateBuilder

if TYPE_CHECKING:
    from nglui.segmentprops import SegmentProperties

DEFAULT_POSTSYN_COLOR = (0.25098039, 0.87843137, 0.81568627)  # CSS3 color turquise
DEFAULT_PRESYN_COLOR = (1.0, 0.38823529, 0.27843137)  # CSS3 color tomato

CONTRAST_CONFIG = {
    "minnie65_phase3_v1": {
        "contrast_controls": True,
        "black": 0.35,
        "white": 0.70,
    },
    "minnie65_public": {
        "contrast_controls": True,
        "black": 0.35,
        "white": 0.70,
    },
}
MAX_URL_LENGTH = 1_750_000
DEFAULT_NGL = fallback_ngl_endpoint


def sort_dataframe_by_root_id(
    df, root_id_column, ascending=False, num_column="n_times", drop=False
):
    """Sort a dataframe so that rows belonging to the same root id are together, ordered by how many times the root id appears.

    Parameters
    ----------
    df : pd.DataFrame
        dataframe to sort
    root_id_column : str
        Column name to use for sorting root ids
    ascending : bool, optional
        Whether to sort ascending (lowest count to highest) or not, by default False
    num_column : str, optional
        Temporary column name to use for count information, by default 'n_times'
    drop : bool, optional
        If True, drop the additional column when returning.

    Returns
    -------
    pd.DataFrame
    """
    cols = df.columns[df.columns != root_id_column]
    if len(cols) == 0:
        df = df.copy()
        df[num_column] = df[root_id_column]
        use_column = num_column
    else:
        use_column = cols[0]

    df[num_column] = df.groupby(root_id_column).transform("count")[use_column]
    if drop:
        return df.sort_values(
            by=[num_column, root_id_column], ascending=ascending
        ).drop(columns=[num_column])
    else:
        return df.sort_values(by=[num_column, root_id_column], ascending=ascending)


def make_line_statebuilder(
    client: CAVEclient,
    point_column_a="pre_pt_position",
    point_column_b="post_pt_position",
    linked_seg_column="pt_root_id",
    description_column=None,
    tag_column=None,
    data_resolution=None,
    group_column=None,
    contrast=None,
    view_kws=None,
    point_layer_name="lines",
    color=None,
    split_positions=False,
):
    """Generate a state builder that puts points on a single column with a linked segmentaton id

    Parameters
    ----------
    client : CAVEclient
        CAVEclient configured for the datastack desired
    point_column_a : str, optional
        column in dataframe to pull points from. Defaults to "pre_pt_position".
    point_column_b : str, optional
        column in dataframe to pull points from. Defaults to "post_pt_position".
    linked_seg_column : str, optional
        column to link to segmentation, None for no column. Defaults to "pt_root_id".
    group_column : str, or list, optional
        column(s) to group annotations by, None for no grouping (default=None)
    tag_column : str, optional
        column to use for tags, None for no tags (default=None)
    description_column : str, optional
        column to use for descriptions, None for no descriptions (default=None)
    contrast : list, optional
        Two elements specifying the black level and white level as
        floats between 0 and 1, by default None. If None, no contrast
        is set.
    view_kws : dict, optional
        dictionary of view keywords to configure neuroglancer view
    split_positions : bool, optional
        whether the position column into x,y,z columns. Defaults to False.
    Returns
    -------
    StateBuilder:
        A statebuilder to make points with linked segmentations
    """
    img_layer, seg_layer = from_client(client, contrast=contrast)
    line_mapper = LineMapper(
        point_column_a=point_column_a,
        point_column_b=point_column_b,
        linked_segmentation_column=linked_seg_column,
        tag_column=tag_column,
        description_column=description_column,
        group_column=group_column,
        split_positions=split_positions,
    )
    ann_layer = AnnotationLayerConfig(
        point_layer_name,
        mapping_rules=[line_mapper],
        linked_segmentation_layer=seg_layer.name,
        data_resolution=data_resolution,
        color=color,
    )
    if view_kws is None:
        view_kws = {}
    return StateBuilder(
        [img_layer, seg_layer, ann_layer],
        client=client,
        view_kws=view_kws,
    )


def make_point_statebuilder(
    client: CAVEclient,
    point_column="pt_position",
    linked_seg_column="pt_root_id",
    data_resolution=None,
    group_column=None,
    tag_column=None,
    description_column=None,
    contrast=None,
    view_kws=None,
    point_layer_name="pts",
    color=None,
    split_positions=False,
):
    """Generate a state builder that puts points on a single column with a linked segmentaton id

    Parameters
    -----------
    client : CAVEclient
        CAVEclient configured for the datastack desired
    point_column : str, optional
        Column in dataframe to pull points from. Defaults to "pt_position".
    linked_seg_column : str, optional
        column to link to segmentation, None for no column. Defaults to "pt_root_id".
    group_column : str, or list, optional
        column(s) to group annotations by, None for no grouping (default=None)
    tag_column : str, optional)
        column to use for tags, None for no tags (default=None)
    description_column : str, optional
        column to use for descriptions, None for no descriptions (default=None)
    contrast : list, optional
        Two elements specifying the black level and white level as
        floats between 0 and 1, by default None. If None, no contrast
        is set.
    view_kws : dict, optional
        dictionary of view keywords to configure neuroglancer view
    split_positions : bool, optional
        whether the position column into x,y,z columns. Defaults to False.

    Returns
    -------
    StateBuilder:
        A statebuilder to make points with linked segmentations
    """
    img_layer, seg_layer = from_client(client, contrast=contrast)
    point_mapper = PointMapper(
        point_column=point_column,
        linked_segmentation_column=linked_seg_column,
        tag_column=tag_column,
        description_column=description_column,
        group_column=group_column,
        split_positions=split_positions,
    )
    ann_layer = AnnotationLayerConfig(
        point_layer_name,
        mapping_rules=[point_mapper],
        linked_segmentation_layer=seg_layer.name,
        data_resolution=data_resolution,
        color=color,
    )
    if view_kws is None:
        view_kws = {}
    return StateBuilder(
        [img_layer, seg_layer, ann_layer],
        client=client,
        view_kws=view_kws,
    )


def make_pre_post_statebuilder(
    client: CAVEclient,
    show_inputs: bool = False,
    show_outputs: bool = False,
    contrast: list = None,
    view_kws: dict = None,
    point_column="ctr_pt_position",
    pre_pt_root_id_col="pre_pt_root_id",
    post_pt_root_id_col="post_pt_root_id",
    dataframe_resolution_pre=None,
    dataframe_resolution_post=None,
    input_layer_name="syns_in",
    output_layer_name="syns_out",
    input_layer_color=DEFAULT_POSTSYN_COLOR,
    output_layer_color=DEFAULT_PRESYN_COLOR,
    split_positions=False,
):
    """Function to generate ChainedStateBuilder with optional pre and post synaptic
    annotation layers

    Parameters
    ----------
    client : CAVEclient
        a CAVEclient configured for datastack to visualize
    show_inputs : bool, optional
        whether to show input synapses. Defaults to False.
    show_outputs : bool, optional
        whether to show output synapses.. Defaults to False.
    contrast : list, optional
        Two elements specifying the black level and white level as
        floats between 0 and 1, by default None. If None, no contrast
        is set.
    view_kws : dict, optional
        view_kws to configure statebuilder, see nglui.StateBuilder. Defaults to None.
        keys are:
            show_slices: Boolean
                sets if slices are shown in the 3d view. Defaults to False.
            layout: str
                `xy-3d`/`xz-3d`/`yz-3d` (sections plus 3d pane), `xy`/`yz`/`xz`/`3d` (only one pane), or `4panel` (all panes). Default is `xy-3d`.
            show_axis_lines: Boolean
                determines if the axis lines are shown in the middle of each view.
            show_scale_bar: Boolean
                toggles showing the scale bar.
            orthographic : Boolean
                toggles orthographic view in the 3d pane.
            position* : 3-element vector
                determines the centered location.
            zoom_image : float
                Zoom level for the imagery in units of nm per voxel. Defaults to 8.
            zoom_3d : float
                Zoom level for the 3d pane. Defaults to 2000. Smaller numbers are more zoomed in.
    point_column : str, optional
        column to pull points for synapses from. Defaults to "ctr_pt_position".
    pre_pt_root_id_col : str, optional
        column to pull pre synaptic ids for synapses from. Defaults to "pre_pt_root_id".
    post_pt_root_id_col : str, optional
        column to pull post synaptic ids for synapses from. Defaults to "post_pt_root_id".
    input_layer_name : str, optional
        name of layer for inputs. Defaults to "syns_in".
    output_layer_name : str, optional
        name of layer for outputs. Defaults to "syns_out".
    split_positions : bool, optional
        whether the position column is split into x,y,z columns. Defaults to False.

    Returns
    -------
    ChainedStateBuilder:
        An instance of a ChainedStateBuilder configured to accept
        a list  starting with None followed by optionally synapse input dataframe
        followed by optionally synapse output dataframe.
    """

    img_layer, seg_layer = from_client(client, contrast=contrast)
    seg_layer.add_selection_map(selected_ids_column="root_id")

    if view_kws is None:
        view_kws = {}

    sb1 = StateBuilder(layers=[img_layer, seg_layer], client=client, view_kws=view_kws)
    state_builders = [sb1]
    if show_inputs:
        # First state builder
        input_point_mapper = PointMapper(
            point_column=point_column,
            linked_segmentation_column=pre_pt_root_id_col,
            split_positions=split_positions,
        )
        inputs_lay = AnnotationLayerConfig(
            input_layer_name,
            mapping_rules=[input_point_mapper],
            linked_segmentation_layer=seg_layer.name,
            data_resolution=dataframe_resolution_post,
            color=input_layer_color,
        )
        sb_in = StateBuilder([inputs_lay], client=client)
        state_builders.append(sb_in)
    if show_outputs:
        output_point_mapper = PointMapper(
            point_column=point_column,
            linked_segmentation_column=post_pt_root_id_col,
            split_positions=split_positions,
        )
        outputs_lay = AnnotationLayerConfig(
            output_layer_name,
            mapping_rules=[output_point_mapper],
            linked_segmentation_layer=seg_layer.name,
            data_resolution=dataframe_resolution_pre,
            color=output_layer_color,
        )
        sb_out = StateBuilder([outputs_lay], client=client)
        state_builders.append(sb_out)
    return ChainedStateBuilder(state_builders)


def make_state_url(df, sb, client, ngl_url=None, target_site=None):
    """Generate a url from a neuroglancer state via a state server.

    Parameters
    ----------
    df : pandas.DataFrame
        Dataframe to pass through statebuilder
    sb : statebuilder.StateBuilder
        Statebuilder to use to render data for link
    client : CAVEclient
        CAVEclient configured with a state server
    ngl_url : str, optional
        Neuroglancer deployment URL, by default None
    target_site : str, optional
        Type of neuroglancer deployment to build link for, by default None.
        This value overrides automatic checking based on the provided url.
        Use `seunglab` for a Seung-lab branch and either `mainline` or `cave-explorer` for the Cave Explorer or main Google branch.

    Returns
    -------
    str
        Url to the uploaded neuroglancer state.
    """
    state = sb.render_state(df, return_as="dict", target_site=target_site)
    state_id = client.state.upload_state_json(state)
    if ngl_url is None:
        ngl_url = client.info.viewer_site()
        if ngl_url is None:
            ngl_url = DEFAULT_NGL
    url = client.state.build_neuroglancer_url(
        state_id, ngl_url=ngl_url, target_site=target_site
    )
    return url


def make_url_robust(
    df: pd.DataFrame,
    sb: StateBuilder,
    client: CAVEclient,
    shorten: str = "if_long",
    ngl_url: str = None,
    max_url_length=MAX_URL_LENGTH,
    target_site=None,
):
    """Generate a url from a neuroglancer state. If too long, return through state server,
    othewise return a url containing the data directly.

    Parameters
    ----------
    df : pandas.DataFrame
        Dataframe to pass through statebuilder
    sb : statebuilder.StateBuilder
        Statebuilder to use to render data for link
    client : CAVEclient
        CAVEclient configured with a state server
    shorten : str, optional
        How to shorten link. one of 'if_long', 'always', 'never'.
        'if_long' will use the state server to shorten links longer than nglui.statebuilder.MAX_URL_LENGTH
        (set to 1,750,000).
        'always' will always use the state server to shorten the url
        'never' will always return the full url.  Defaults to "if_long".
    ngl_url : str, optional
        Neuroglancer deployment URL, by default None
    max_url_length : int, optional
        Maximum length of url to return directly, by default 1_750_000
    target_site : str, optional
        Type of neuroglancer deployment to build link for, by default None.
        This value overrides automatic checking based on the provided url.
        Use `seunglab` for a Seung-lab branch and either `mainline` or `cave-explorer` for the Cave Explorer or main Google branch.

    Returns
    -------
    str
        URL containing the state created by the statebuilder.
    """

    if shorten == "if_long":
        url = sb.render_state(
            df, return_as="url", url_prefix=ngl_url, target_site=target_site
        )
        if len(url) > max_url_length:
            url = make_state_url(
                df, sb, client, ngl_url=ngl_url, target_site=target_site
            )
    elif shorten == "always":
        url = make_state_url(df, sb, client, ngl_url=ngl_url, target_site=target_site)
    elif shorten == "never":
        url = sb.render_state(
            df, return_as="url", url_prefix=ngl_url, target_site=target_site
        )
    else:
        raise (ValueError('shorten should be one of ["if_long", "always", "never"]'))
    return url


def package_state(
    df: pd.DataFrame,
    sb: StateBuilder,
    client: CAVEclient,
    shorten: str = "if_long",
    return_as: str = "url",
    ngl_url: str = None,
    link_text: str = "Neuroglancer Link",
    target_site: str = None,
):
    """Automate creating a state from a statebuilder and
    a dataframe, return it in the desired format, shortening if desired.

    Parameters
    ----------
    df : pandas.DataFrame
        Dataframe to pass through statebuilder
    sb : statebuilder.StateBuilder
        Statebuilder to use to render data for link
    client : CAVEclient
        CAVEclient configured with a state server
    shorten : str, optional
        How to shorten link. one of 'if_long', 'always', 'never'.
        'if_long' will use the state server to shorten links longer than nglui.statebuilder.MAX_URL_LENGTH
        (set to 1,750,000).
        'always' will always use the state server to shorten the url
        'never' will always return the full url.  Defaults to "if_long".
    return_as : str, optional
        How to return the state. one of 'html', 'url', 'json'.
        'html' will return an ipython clickable link
        'url' will return a string with the url
        'json' will return the state as a dictionary
    ngl_url : str, optional
        Neuroglancer deployment URL, by default None
    link_text : str, optional
        Text to use for the link, by default "Neuroglancer Link"
    target_site : str, optional
        Type of neuroglancer deployment to build link for, by default None.
        This value overrides automatic checking based on the provided url.
        Use `seunglab` for a Seung-lab branch and either `mainline` or `cave-explorer` for the Cave Explorer or main Google branch.

    Returns
    -------
    HTML, str or dict
        state in format specified by return_as
    """
    if ngl_url is None:
        ngl_url = client.info.viewer_site()
        if ngl_url is None:
            ngl_url = DEFAULT_NGL

    if (return_as == "html") or (return_as == "url"):
        url = make_url_robust(
            df, sb, client, shorten=shorten, ngl_url=ngl_url, target_site=target_site
        )
        if return_as == "html":
            return HTML(f'<a href="{url}">{link_text}</a>')
        else:
            return url
    elif return_as == "json":
        return sb.render_state(
            df, return_as=return_as, ngl_url=ngl_url, target_site=target_site
        )
    else:
        raise (
            ValueError(
                f'return_as={return_as} not a valid choice, choose one of "html", "url", or "json")'
            )
        )


def make_synapse_neuroglancer_link(
    synapse_df,
    client,
    return_as="html",
    shorten="always",
    contrast=None,
    point_column="ctr_pt_position",
    dataframe_resolution=None,
    group_connections=True,
    link_pre_and_post=True,
    ngl_url=None,
    view_kws=None,
    pre_post_columns=None,
    neuroglancer_link_text="Neuroglancer Link",
    color=None,
    split_positions=False,
    target_site=None,
):
    """Generate a neuroglancer link from a synapse dataframe as returned from CAVEclient.materialize.synapse_query.

    Parameters
    ----------
    synapse_df : pandas.DataFrame
        DataFrame where each row represents a synapse.
    client : caveclient.CAVEclient
        A CAVEclient instance.
    return_as : str, optional
        How to return the URL. Valid options are:
        - 'html': Returns an IPython HTML element.
        - 'url': Returns a URL string.
        - 'json': Returns a dictionary representing the Neuroglancer state.
        Defaults to 'html'.
    shorten : str, optional
        Whether to shorten the link. Valid options are:
        - 'always': Always shorten the link.
        - 'if_long': Shorten the link if it exceeds MAX_URL_LENGTH.
        - 'never': Never shorten the link.
        Defaults to 'always'.
    contrast : list, optional
        List of two floats between 0 and 1, specifying the black and white levels.
        If None, no contrast is set.
    point_column : str, optional
        Column in the DataFrame containing synapse positions. Defaults to 'ctr_pt_position'.
    dataframe_resolution : list, optional
        List of length 3, specifying the resolution units of the position column.
        If None, attempts to get the resolution from DataFrame metadata or client.info.viewer_resolution().
    group_connections : bool, optional
        Whether to group synapses within the same connection (between the same neurons).
        Defaults to True.
    link_pre_and_post : bool, optional
        Whether to link the synapse annotations to the pre- and post-synaptic partners.
        Defaults to True.
    ngl_url : str, optional
        URL of the Neuroglancer instance to use. Defaults to client.info.viewer_site() or DEFAULT_NGL.
    view_kws : dict, optional
        Dictionary containing viewer settings for Neuroglancer. Available keys:
        - show_slices: Boolean, sets if slices are shown in the 3D view.
        - layout: str, specifies the viewer layout (e.g., 'xy-3d', '4panel').
        - show_axis_lines: Boolean, determines if axis lines are shown.
        - show_scale_bar: Boolean, toggles the scale bar.
        - orthographic: Boolean, toggles orthographic view in the 3D pane.
        - position: 3-element vector, sets the centered location.
        - zoom_image: float, zoom level for the imagery in nm per voxel.
        - zoom_3d: float, zoom level for the 3D pane.
    pre_post_columns : list, optional
        List of two strings, specifying the column names for pre- and post-synaptic root_ids.
        Defaults to ['pre_pt_root_id', 'post_pt_root_id'].
    neuroglancer_link_text : str, optional
        Text to use in the returned HTML link. Defaults to 'Neuroglancer Link'.
    color : list(float) or str, optional
        Color of synapse points as an RGB list [0, 1], hex string, or common name.
    split_positions : bool, optional
        Whether the position column is split into x, y, and z columns.
    target_site : str, optional
        Type of neuroglancer deployment to build link for, by default None.
        This value overrides automatic checking based on the provided url.
        Use `seunglab` for a Seung-lab branch and either `mainline` or `cave-explorer` for the Cave Explorer or main Google branch.

    Returns
    -------
    IPython.HTML, str, or dict
        Representation of the Neuroglancer state, depending on the `return_as` parameter.

    Raises
    ------
    ValueError
        If the specified `point_column` is not found in the DataFrame.
    """
    if point_column not in synapse_df.columns:
        raise ValueError(f"point_column={point_column} not in dataframe")
    if pre_post_columns is None:
        pre_post_columns = ["pre_pt_root_id", "post_pt_root_id"]
    if dataframe_resolution is None:
        dataframe_resolution = synapse_df.attrs.get("dataframe_resolution", None)

    if group_connections:
        group_column = pre_post_columns
    else:
        group_column = None
    if link_pre_and_post:
        linked_columns = pre_post_columns
    else:
        linked_columns = None

    sb = make_point_statebuilder(
        client,
        point_column=point_column,
        linked_seg_column=linked_columns,
        group_column=group_column,
        data_resolution=dataframe_resolution,
        contrast=contrast,
        view_kws=view_kws,
        point_layer_name="synapses",
        color=color,
        split_positions=split_positions,
    )
    return package_state(
        synapse_df,
        sb,
        client,
        shorten,
        return_as,
        ngl_url,
        neuroglancer_link_text,
        target_site=target_site,
    )


def make_neuron_neuroglancer_link(
    client,
    root_ids,
    return_as="html",
    shorten="always",
    show_inputs=False,
    show_outputs=False,
    sort_inputs=True,
    sort_outputs=True,
    sort_ascending=False,
    input_color=DEFAULT_POSTSYN_COLOR,
    output_color=DEFAULT_PRESYN_COLOR,
    contrast=None,
    timestamp=None,
    view_kws=None,
    point_column="ctr_pt_position",
    pre_pt_root_id_col="pre_pt_root_id",
    post_pt_root_id_col="post_pt_root_id",
    input_layer_name="syns_in",
    output_layer_name="syns_out",
    ngl_url=None,
    link_text="Neuroglancer Link",
    target_site=None,
):
    """function to create a neuroglancer link view of a neuron, optionally including inputs and outputs

    Parameters
    ----------
    client : CAVEclient
        A CAVEclient configured for the datastack to visualize.
    root_ids : Iterable[int]
        The root_ids to build the visualization around.
    return_as : str, optional
        How to return the URL or state. Valid options are:
        - 'html': Returns an IPython HTML element.
        - 'json': Returns a dictionary representing the Neuroglancer state.
        - 'url': Returns a URL string.
        Defaults to 'html'.
    shorten : str, optional
        Whether to shorten the link. Valid options are:
        - 'always': Always shorten the link.
        - 'if_long': Shorten the link if it exceeds MAX_URL_LENGTH.
        - 'never': Never shorten the link.
        Defaults to 'if_long'.
    show_inputs : bool, optional
        Whether to include input synapses. Defaults to False.
    show_outputs : bool, optional
        Whether to include output synapses. Defaults to False.
    sort_inputs : bool, optional
        Whether to sort input synapses by presynaptic root id, ordered by synapse count.
        Defaults to True.
    sort_outputs : bool, optional
        Whether to sort output synapses by presynaptic root id, ordered by postsynaptic synapse count.
        Defaults to True.
    sort_ascending : bool, optional
        If sorting, whether to sort ascending (lowest synapse count to highest).
        Defaults to False.
    input_color : list(float) or str, optional
        Color of input synapse points as an RGB list [0, 1], hex string, or common name.
    output_color : list(float) or str, optional
        Color of output synapse points as an RGB list [0, 1], hex string, or common name.
    contrast : list, optional
        List of two floats between 0 and 1, specifying the black and white levels.
        If None, no contrast is set.
    timestamp : datetime.datetime, optional
        Timestamp to use for the query. Defaults to None, which will use the materialized version.
    view_kws : dict, optional
        Dictionary containing viewer settings for Neuroglancer. See nglui.StateBuilder for options.
        See the previous docstring for details on available keys.
    point_column : str, optional
        Column to pull synapse positions from. Defaults to "ctr_pt_position".
    pre_pt_root_id_col : str, optional
        Column to pull presynaptic IDs for synapses from. Defaults to "pre_pt_root_id".
    post_pt_root_id_col : str, optional
        Column to pull postsynaptic IDs for synapses from. Defaults to "post_pt_root_id".
    input_layer_name : str, optional
        Name of the layer for input synapses. Defaults to "syns_in".
    output_layer_name : str, optional
        Name of the layer for output synapses. Defaults to "syns_out".
    ngl_url : str, optional
        URL of the Neuroglancer instance to use. Defaults to the default viewer set in the datastack.
    link_text : str, optional
        Text to use for the HTML return. Defaults to 'Neuroglancer Link'.

    Raises
    ------
    ValueError
        If the specified point column is not present in the synapse table.

    Returns
    -------
    IPython.HTML, str, or dict
        Representation of the Neuroglancer state, depending on the `return_as` parameter.
    """
    if not isinstance(root_ids, Iterable):
        root_ids = [root_ids]
    df1 = pd.DataFrame({"root_id": root_ids})
    dataframes = [df1]
    data_resolution_pre = None
    data_resolution_post = None
    if show_inputs:
        syn_in_df = client.materialize.synapse_query(
            post_ids=root_ids,
            timestamp=timestamp,
            desired_resolution=client.info.viewer_resolution(),
            split_positions=True,
        )
        data_resolution_pre = syn_in_df.attrs["dataframe_resolution"]
        if sort_inputs:
            syn_in_df = sort_dataframe_by_root_id(
                syn_in_df, pre_pt_root_id_col, ascending=sort_ascending, drop=True
            )
        dataframes.append(syn_in_df)
    if show_outputs:
        syn_out_df = client.materialize.synapse_query(
            pre_ids=root_ids,
            timestamp=timestamp,
            desired_resolution=client.info.viewer_resolution(),
            split_positions=True,
        )
        data_resolution_post = syn_out_df.attrs["dataframe_resolution"]
        if sort_outputs:
            syn_out_df = sort_dataframe_by_root_id(
                syn_out_df, post_pt_root_id_col, ascending=sort_ascending, drop=True
            )
        dataframes.append(syn_out_df)
    sb = make_pre_post_statebuilder(
        client,
        show_inputs=show_inputs,
        show_outputs=show_outputs,
        contrast=contrast,
        point_column=point_column,
        view_kws=view_kws,
        pre_pt_root_id_col=pre_pt_root_id_col,
        post_pt_root_id_col=post_pt_root_id_col,
        input_layer_name=input_layer_name,
        output_layer_name=output_layer_name,
        input_layer_color=input_color,
        output_layer_color=output_color,
        dataframe_resolution_pre=data_resolution_pre,
        dataframe_resolution_post=data_resolution_post,
        split_positions=True,
    )
    return package_state(
        dataframes,
        sb,
        client,
        shorten,
        return_as,
        ngl_url,
        link_text,
        target_site=target_site,
    )


def from_client(
    client,
    image_name=None,
    segmentation_name=None,
    contrast=None,
    use_skeleton_service=False,
):
    """Generate basic image and segmentation layers from a FrameworkClient

    Parameters
    ----------
    client : caveclient.CAVEclient
        A CAVEclient with a specified datastack
    image_name : str, optional
        Name for the image layer, by default None.
    segmentation_name : str, optional
        Name for the segmentation layer, by default None
    contrast : list-like or False, optional
        Two elements specifying the black level and white level as
        floats between 0 and 1, by default None. If None, no contrast
        is set.
    use_skeleton_service : bool, optional
        If True, uses a skeleton service, if advertised, with the segmentation. Defaults to False.

    Returns
    -------
    image_layer : ImageLayerConfig
        Image layer with default values from the client
    seg_layer : ImageLayerConfig
        Segmentation layer with default values from the client
    """
    if contrast is None:
        config = CONTRAST_CONFIG.get(
            client.datastack_name, {"contrast_controls": True, "black": 0, "white": 1}
        )
    elif contrast is False:
        config = {}
    else:
        config = {"contrast_controls": True, "black": contrast[0], "white": contrast[1]}

    if image_name is not False:
        img_layer = ImageLayerConfig(
            client.info.image_source(), name=image_name, **config
        )
    else:
        img_layer = None
    if segmentation_name is not False:
        if use_skeleton_service:
            skeleton_source = client.info.get_datastack_info().get(
                "skeleton_source", None
            )
            if skeleton_source is None:
                warn(
                    "Skeleton source requested but no skeleton source found in datastack info."
                )
        else:
            skeleton_source = None
        seg_layer = SegmentationLayerConfig(
            client.info.segmentation_source(),
            name=segmentation_name,
            skeleton_source=skeleton_source,
        )
    else:
        seg_layer = None
    return img_layer, seg_layer


def segment_property_link(
    seg_props: Union[str, SegmentProperties],
    client: CAVEclient,
    ngl_url: Optional[str] = None,
    return_as: Literal["html", "url", "viewer", "json", "dict", "short"] = "html",
):
    """Returns a basic link to a default neuroglancer state and segment properties.

    Parameters
    ----------
    seg_props : segment_properties.SegmentProperties or str
        A segment propeties object or a URL to a segment properties.
    client : caveclient.CAVEclient
        A caveclient object.
    ngl_url : str, optional
        URL for a neuroglancer deployment, by default None which uses a default Spelunker deployment.
    return_as : str, optional
        Select how the data comes back from statebuilder, by default "html".
        See statebuilder.render_state for more information.

    Returns
    -------
    string or neuroglancer.Viewer
    """
    img, seg = from_client(client)
    if isinstance(seg_props, str):
        seg_prop_url = seg_props
    else:
        prop_id = client.state.upload_property_json(seg_props.to_dict())
        seg_prop_url = client.state.build_neuroglancer_url(
            prop_id,
            target_site="cave-explorer",
            format_properties=True,
        )
    ngl_url = neuroglancer_url(ngl_url, target_site="spelunker")
    seg.add_segment_propeties(seg_prop_url)
    sb = StateBuilder(
        [img, seg], client=client, url_prefix=ngl_url, target_site="cave-explorer"
    )
    return sb.render_state(return_as=return_as, url_prefix=ngl_url)



from typing import Optional

from ..site_utils import get_default_config, is_mainline
from .ev_base import EasyViewerMainline, EasyViewerSeunglab


def EasyViewer(
    target_site: Optional[str] = None,
    config_key: str = "default",
):
    """Factory function to return the appropriate EasyViewer object based on the target site.

    Parameters
    ----------
    target_site : Optional[str], optional
        Type of target site, either "seunglab" or ""spelunker/mainline/"cave-explorer", by default None.
        If None, uses the configuration setting.
    config_key : str, optional
        Name of the configuration to use, by default "default"

    Returns
    -------
    EasyViewer
        EasyViewer object based on the target site to aid in building states.
    """
    if target_site is None:
        target_site = get_default_config(config_key)["target_site"]
    if not is_mainline(target_site):
        return EasyViewerSeunglab()
    elif is_mainline(target_site):
        return EasyViewerMainline()
    else:
        raise ValueError(
            f'Unknown target: {target_site}. Must be one of "seunglab" or "spelunker/mainline"/"cave-explorer"'
        )



from __future__ import annotations

import numbers
from datetime import datetime
from typing import Optional, Union
from warnings import warn

import numpy as np

from ..easyviewer import EasyViewerMainline
from ..segmentprops.base import SegmentProperties
from .mappers import (
    AnnotationMapperBase,
    SelectionMapper,
)

DEFAULT_IMAGE_LAYER = "img"
DEFAULT_SEG_LAYER = "seg"
DEFAULT_ANNO_LAYER = "anno"

DEFAULT_SEGMENTATION_VIEW_KWS = {
    "alpha_selected": 0.3,
    "alpha_3d": 1,
    "alpha_unselected": 0,
}


class LayerConfigBase:
    """Base class for configuring layers

    Parameters
    ----------
    name : str
        Layer name for reference and display
    type : str
        Layer type. Usually handled by the subclass
    source : str
        Datasource for the layer
    color : str
        Hex string (with starting hash).
    active : bool,
        If True, becomes a selected layer.
    """

    def __init__(
        self,
        name,
        type,
        source,
        color,
        active,
    ):
        self._config = dict(
            type=type,
            name=str(name),
            source=source,
            color=color,
            active=active,
        )

    @property
    def type(self):
        return self._config.get("type", None)

    @property
    def name(self):
        return self._config.get("name", None)

    @name.setter
    def name(self, n):
        self._config["name"] = n

    @property
    def source(self):
        return self._config.get("source", None)

    @source.setter
    def source(self, s):
        self._config["source"] = s

    @property
    def color(self):
        return self._config.get("color", None)

    @color.setter
    def color(self, c):
        self._config["color"] = c

    @property
    def active(self):
        return self._config.get("active", False)

    @active.setter
    def active(self, val):
        self._config["active"] = val

    def _render_layer(
        self, viewer, data, viewer_resolution=None, return_annos=False, client=None
    ):
        """Applies rendering rules"""
        annos = self._specific_rendering(
            viewer,
            data,
            viewer_resolution=viewer_resolution,
            return_annos=return_annos,
            client=client,
        )
        if self.active:
            viewer.set_selected_layer(self.name)
        if return_annos:
            return annos
        else:
            pass

    def _add_layer(self, viewer):
        "Subclass implements layer addition rules"

    def _set_view_options(self, *args, **kwargs):
        "Subclass implements own rules"

    def _specific_rendering(
        self,
        viewer,
        data,
        viewer_resolution=None,
        return_annos=False,
        client=None,
    ):
        """Subclasses implement specific rendering rules"""


class ImageLayerConfig(LayerConfigBase):
    """Image layer configuration class.

    This provides the rules for setting up an image layer in neuroglancer.

    Parameters
    ----------
    source : str
        Cloudpath to an image source
    name : str, optional
        Name of the image layer. By default, 'img'.
    active : bool, optional
        If True, makes the layer active in neuroglancer. Default is False.
    contrast_controls : bool, optional
        If True, gives the layer a user-controllable brightness and contrast shader. Default is False.
    black : float, optional
        If contrast_controls is True, sets the default black level. Default is 0.0.
    white : float, optional
        If contrast_controls is True, sets the default white level. Default is 1.0.
    """

    def __init__(
        self,
        source,
        name=None,
        active=False,
        contrast_controls=False,
        black=0.0,
        white=1.0,
    ):
        if name is None:
            name = DEFAULT_IMAGE_LAYER
        super(ImageLayerConfig, self).__init__(
            name=name, type="image", source=source, color=None, active=active
        )
        self._contrast_controls = contrast_controls
        self._black = black
        self._white = white

    def _add_layer(self, viewer):
        viewer.add_image_layer(self.name, self.source)

    def _specific_rendering(
        self,
        viewer,
        data,
        viewer_resolution=None,
        return_annos=False,
        client=None,
    ):
        if self._contrast_controls:
            viewer.add_contrast_shader(self.name, black=self._black, white=self._white)


class SegmentationLayerConfig(LayerConfigBase):
    """Configuration class for segmentation layers

    Parameters
    ----------
    source : str
        Segmentation source
    name : str, optional,
        Layer name.
    selected_ids_column : str or list-like, optional.
        Column name (or list of column names) to use for selected ids.
    fixed_ids : list-like, optional.
        List of root ids to select directly.
    fixed_id_colors : list-like, optional.
        List of colors for fixed ids. Should be the same length as fixed_ids, although null entries can be padded with None values.
    color_column : str, optional.
        # at the begining.
        Column to use for color values for selected objects. Values should be RGB hex strings with a
    active : bool, optional.
        If True, makes the layer selected. Default is False.
    alpha_selected: float, optional
        Opacity of selected segmentations in the image layer. Optional, default is 0.3.
    alpha_3d: float, optional
        Opacity of meshes. Optional, default is 1.
    alpha_unselected: float, optional
        Opacity of unselected segments. Optional, default is 0.
    split_point_map: mappers.SplitPointMap or None, optional
        If not None, provides an object to map the dataframe input to multicut points. Default is None.
    view_kws : dict, optional.
        Keyword arguments for viewer.set_segmentation_view_options. Sets selected (and unselected) segmetation alpha values. Defaults to values in DEFAULT_SEGMENTATION_VIEW_KWS dict specified in this module.
    timestamp : float or datetime, optional.
        Timestamp at which to fix the chunkedgraph in either unix epoch or datetime format. Optional, default is None.
    mapping_set : str, optional.
        Name of the mapping set, the key in the data dictionary for statebuilder. Optional, default is None.
    segment_properties: str, optional
        Location of a segment properties file. Optional, default is None.
    skeleton_source: str, optional
        Location of a skeleton source. Optional, default is None.
    skeleton_shader: str, optional
        Shader function for rendering skeletons, if set with `skeleton_source`. Optional, default is None.
    """

    def __init__(
        self,
        source,
        name=None,
        selected_ids_column=None,
        fixed_ids=None,
        fixed_id_colors=None,
        color_column=None,
        active=False,
        alpha_selected=0.3,
        alpha_3d=1,
        alpha_unselected=0,
        split_point_map=None,
        view_kws=None,
        timestamp=None,
        data_resolution=None,
        mapping_set=None,
        segment_properties=None,
        skeleton_source=None,
        skeleton_shader=None,
    ):
        if name is None:
            name = DEFAULT_SEG_LAYER

        super(SegmentationLayerConfig, self).__init__(
            name=name, type="segmentation", source=source, color=None, active=active
        )
        self._config["data_resolution"] = data_resolution
        self._config["segment_properties"] = segment_properties
        self._config["skeleton_source"] = skeleton_source
        self._config["skeleton_shader"] = skeleton_shader

        if selected_ids_column is not None or fixed_ids is not None:
            self._selection_map = SelectionMapper(
                data_columns=selected_ids_column,
                fixed_ids=fixed_ids,
                fixed_id_colors=fixed_id_colors,
                color_column=color_column,
                mapping_set=mapping_set,
            )
        else:
            self._selection_map = None

        self._split_point_map = split_point_map

        if isinstance(timestamp, datetime):
            timestamp = timestamp.timestamp()
        self.timestamp = timestamp

        base_seg_kws = DEFAULT_SEGMENTATION_VIEW_KWS.copy()
        if view_kws is None:
            view_kws = {
                "alpha_selected": alpha_selected,
                "alpha_3d": alpha_3d,
                "alpha_unselected": alpha_unselected,
            }
        base_seg_kws.update(view_kws)
        self._view_kws = view_kws
        self._segment_property_map = None

    @property
    def data_resolution(self):
        return self._config.get("data_resolution", None)

    @property
    def segment_properties(self):
        if isinstance(self._config.get("segment_properties"), str):
            return [self._config.get("segment_properties")]
        else:
            return self._config.get("segment_properties", [])

    @property
    def source(self):
        if self.segment_properties:
            seg_prop_url = self.segment_properties
        else:
            seg_prop_url = []
        if self._config.get("source"):
            if isinstance(self._config.get("source"), str):
                source = [self._config.get("source")]
            else:
                source = self._config.get("source")
        else:
            source = []
        out = source + seg_prop_url
        if len(out) == 1:
            return out[0]
        else:
            return out

    @property
    def skeleton_source(self):
        return self._config.get("skeleton_source", None)

    @property
    def skeleton_shader(self):
        return self._config.get("skeleton_shader", None)

    def add_segment_propeties(
        self,
        segment_property: str,
    ):
        self._config["segment_properties"] = segment_property

    def add_segment_properties_map(
        self,
        id_col: str = "pt_root_id",
        label_col: Optional[Union[str, list[str]]] = None,
        description_col: Optional[str] = None,
        string_cols: Optional[Union[str, list[str]]] = None,
        number_cols: Optional[Union[str, list[str]]] = None,
        tag_value_cols: Optional[Union[str, list[str]]] = None,
        tag_bool_cols: Optional[list[str]] = None,
        tag_descriptions: Optional[dict] = None,
        mapping_set: Optional[str] = None,
        allow_disambiguation: Optional[bool] = True,
        label_separator: Optional[str] = "_",
        label_format_map: Optional[str] = None,
    ):
        segment_property_map = {
            "id_col": id_col,
            "label_col": label_col,
            "description_col": description_col,
            "string_cols": string_cols,
            "number_cols": number_cols,
            "tag_value_cols": tag_value_cols,
            "tag_bool_cols": tag_bool_cols,
            "tag_descriptions": tag_descriptions,
            "allow_disambiguation": allow_disambiguation,
            "label_separator": label_separator,
            "label_format_map": label_format_map,
        }
        if mapping_set:
            if not isinstance(self._segment_property_map, dict):
                self._segment_property_map = {}
            self._segment_property_map[mapping_set] = segment_property_map
        else:
            self._segment_property_map = segment_property_map

    def _render_single_segment_prop(self, data, seg_prop_map, client):
        if data is None:
            return ""
        props = SegmentProperties.from_dataframe(data, **seg_prop_map)
        pid = client.state.upload_property_json(props.to_dict())
        return client.state.build_neuroglancer_url(
            pid,
            target_site="cave-explorer",
            format_properties=True,
        )

    def _render_segment_property_map(self, data, client, target_site=None):
        if target_site == "seunglab":
            warn("Cannot render segment properties for seunglab")
            return None
        if client is None:
            raise ValueError(
                "Client must be provided to dynamically render segment properties"
            )
        seg_prop_urls = []
        if isinstance(data, dict):
            for k, v in data.items():
                if k in self._segment_property_map:
                    seg_prop_urls.append(
                        self._render_single_segment_prop(
                            v, self._segment_property_map.get(k), client
                        )
                    )
        else:
            seg_prop_urls.append(
                self._render_single_segment_prop(
                    data, self._segment_property_map, client
                )
            )
        return seg_prop_urls

    def add_selection_map(
        self,
        selected_ids_column=None,
        fixed_ids=None,
        fixed_id_colors=None,
        color_column=None,
        mapping_set=None,
    ):
        """Add rules for selecting active segment ids and their colors

        Parameters
        ----------
        selected_ids_column : str, optional
            Dataframe column to use for adding selected segment ids to the segmentation layer, by default None
        fixed_ids : int or list, optional
            Add one or more segment ids to be active, independent of the data, by default None
        fixed_id_colors : list, optional
            Add a list of colors (hex, rgb, or CSS3 string) to assign to the fixed ids, by default None
        color_column : str, optional
            Dataframe column to use for adding selected segment colors, by default None
        mapping_set : str, optional
            Name of the mapping set, the key in the data dictionary for statebuilder, by default None
        """
        if self._selection_map is not None:
            if isinstance(selected_ids_column, str):
                selected_ids_column = [selected_ids_column]
            if selected_ids_column is not None:
                data_columns = self._selection_map.data_columns + selected_ids_column
            else:
                data_columns = self._selection_map.data_columns
            if fixed_ids is not None:
                new_fixed_ids = np.concatenate(
                    (self._selection_map.fixed_ids.tolist(), np.atleast_1d(fixed_ids))
                ).tolist()
                old_fixed_id_colors = self._selection_map.fixed_id_colors
                if len(old_fixed_id_colors) > len(
                    self._selection_map.fixed_ids.tolist()
                ):
                    old_fixed_id_colors = old_fixed_id_colors[
                        : len(len(self._selection_map.fixed_ids.tolist()))
                    ]
                else:
                    while len(old_fixed_id_colors) < len(
                        self._selection_map.fixed_ids.tolist()
                    ):
                        old_fixed_id_colors += [None]
                if fixed_id_colors is None:
                    fixed_id_colors = len(fixed_ids) * [None]
                elif isinstance(fixed_id_colors, str) or isinstance(
                    fixed_ids, numbers.Number
                ):
                    fixed_id_colors = [fixed_id_colors]
                new_fixed_id_colors = old_fixed_id_colors + fixed_id_colors
            else:
                new_fixed_ids = self._selection_map.fixed_ids
                new_fixed_id_colors = self._selection_map.fixed_id_colors
            if color_column is None:
                color_column = self._selection_map.color_column
        else:
            data_columns = selected_ids_column
            new_fixed_ids = fixed_ids
            new_fixed_id_colors = fixed_id_colors

        self._selection_map = SelectionMapper(
            data_columns=data_columns,
            fixed_ids=new_fixed_ids,
            fixed_id_colors=new_fixed_id_colors,
            color_column=color_column,
            mapping_set=mapping_set,
        )

    def _add_layer(self, viewer):
        viewer.add_segmentation_layer(self.name, self.source)
        if self.skeleton_source is not None:
            viewer.add_skeleton_source(
                self.name, self.skeleton_source, self.skeleton_shader
            )

    def _specific_rendering(
        self,
        viewer,
        data,
        viewer_resolution=None,
        return_annos=False,
        client=None,
    ):
        if self._selection_map is not None:
            selected_ids = self._selection_map.selected_ids(data)
            colors = self._selection_map.seg_colors(data)
            viewer.add_selected_objects(self.name, selected_ids, colors)

        if self._segment_property_map is not None:
            if isinstance(viewer, EasyViewerMainline):
                target_site = "cave-explorer"
            else:
                target_site = "seunglab"
            seg_prop_urls = self._render_segment_property_map(data, client, target_site)
            for seg_prop_url in seg_prop_urls:
                viewer.append_source_to_segmentation_layer(self.name, seg_prop_url)

        if self._split_point_map is not None:
            (
                seg_id,
                points_red,
                points_blue,
                sv_red,
                sv_blue,
            ) = self._split_point_map._render_data(
                data,
                data_resolution=self.data_resolution,
                viewer_resolution=viewer_resolution,
                viewer=viewer,
            )
            viewer.set_multicut_points(
                self.name,
                seg_id,
                points_red,
                points_blue,
                sv_red,
                sv_blue,
                self._split_point_map.focus,
            )

        viewer.set_timestamp(self.name, self.timestamp)
        viewer.set_segmentation_view_options(self.name, **self._view_kws)


class AnnotationLayerConfig(LayerConfigBase):
    """Configuration class for annotation layers

    Parameters
    ----------
    name : str, optional
        Layer name. By default, 'annos'
    color : str, optional
        Hex color code with an initial #. By default, None
    linked_segmentation_layer : str, optional
        Name of a linked segmentation layer for selected ids. By default, None
    mapping_rules : PointMapper, LineMapper, SphereMapper or list, optional
        One rule or a list of rules mapping data to annotations. By default, []
    array_data : bool, optional
        If True, allows simple mapping where one or more arrays are passed instead of a dataframe.
        Only allows basic annotation creation, no tags, linked segmentations, or other rich features.
    tags : list, optional
        List of tags for the layer.
    active : bool, optional
        If True, makes the layer selected. Default is True (unlike for image/segmentation layers).
    """

    def __init__(
        self,
        name=None,
        color=None,
        linked_segmentation_layer=None,
        mapping_rules=[],
        array_data=False,
        tags=None,
        active=True,
        filter_by_segmentation=False,
        brackets_show_segmentation=True,
        selection_shows_segmentation=True,
        filter_query=None,
        data_resolution=None,
    ):
        if name is None:
            name = DEFAULT_ANNO_LAYER

        super(AnnotationLayerConfig, self).__init__(
            name=name, type="annotation", color=color, source=None, active=active
        )
        self._config["linked_segmentation_layer"] = linked_segmentation_layer
        self._config["filter_by_segmentation"] = filter_by_segmentation
        self._config["selection_shows_segmentation"] = selection_shows_segmentation
        self._config["brackets_show_segmentation"] = brackets_show_segmentation
        self._config["filter_query"] = filter_query
        self._config["data_resolution"] = data_resolution

        if issubclass(type(mapping_rules), AnnotationMapperBase):
            mapping_rules = [mapping_rules]
        if array_data is True:
            if len(mapping_rules) > 1:
                raise ValueError("Only one mapping rule can be set using array data")
            for mr in mapping_rules:
                mr.array_data = array_data
        self._array_data = array_data
        self._annotation_map_rules = mapping_rules
        self._tags = tags

    @property
    def linked_segmentation_layer(self):
        return self._config.get("linked_segmentation_layer", None)

    @property
    def filter_by_segmentation(self):
        return self._config.get("filter_by_segmentation", None)

    @property
    def selection_shows_segmentation(self):
        return self._config.get("selection_shows_segmentation", None)

    @property
    def brackets_show_segmentation(self):
        return self._config.get("brackets_show_segmentation", None)

    @property
    def filter_query(self):
        return self._config.get("filter_query", None)

    @property
    def data_resolution(self):
        return self._config.get("data_resolution", None)

    def _add_layer(self, viewer):
        viewer.add_annotation_layer(
            self.name,
            color=self.color,
            linked_segmentation_layer=self.linked_segmentation_layer,
            filter_by_segmentation=self.filter_by_segmentation,
            selection_shows_segmentation=self.selection_shows_segmentation,
            brackets_show_segmentation=self.brackets_show_segmentation,
            tags=self._tags,
        )

    def _set_view_options(self, viewer, data, viewer_resolution=None):
        pos = None
        if data is not None:
            if self.filter_query is not None:
                data = data.query(self.filter_query)
            for rule in self._annotation_map_rules:
                pos = rule._get_position(
                    data,
                    data_resolution=self.data_resolution,
                    viewer_resolution=viewer_resolution,
                )
                viewer.set_view_options(position=pos)
                if pos is not None:
                    break
        return pos

    def _specific_rendering(
        self,
        viewer,
        data,
        viewer_resolution=None,
        return_annos=False,
        client=None,
    ):
        annos = []
        for rule in self._annotation_map_rules:
            rule.tags = self._tags
            if data is not None:
                if self.filter_query is not None:
                    data = data.query(self.filter_query)
                if len(data) > 0:
                    annos.extend(
                        rule._render_data(
                            data,
                            data_resolution=self.data_resolution,
                            viewer_resolution=viewer_resolution,
                            viewer=viewer,
                        )
                    )
        if return_annos:
            return annos
        else:
            viewer.add_annotations(self.name, annos)



from functools import partial
from itertools import chain
from typing import List, Optional, Union

import attrs
import numpy as np
import pandas as pd

"""
Options and validation for neuroglancer segment properties based on the segment properties spec:
https://github.com/google/neuroglancer/blob/master/src/datasource/precomputed/segment_properties.md
"""


class SegmentPropertyBase:
    pass


ALLOWED_NUMBER_DATA_TYPES = [
    "uint8",
    "int8",
    "uint16",
    "int16",
    "uint32",
    "int32",
    "float32",
]
NONE_ATTR_FILTERS = ["tag_descriptions", "description"]


def list_of_strings(x: list) -> list:
    return [str(v) for v in x]


def space_to_underscore(x: list) -> list:
    return [str(y).replace(" ", "_") for y in x]


def sort_tag_arrays(x: list) -> list:
    return [sorted(y) for y in x]


def zero_null_strings(x: list) -> list:
    return [y if not is_null_value(y) else "" for y in x]


def is_null_value(value):
    if value is None or pd.isna(value) or value == "":
        return True
    else:
        return False


def preprocess_string_column(x: list) -> list:
    return space_to_underscore(zero_null_strings(x))


@attrs.define
class InlineProperties:
    ids = attrs.field(type=List[int], converter=list_of_strings, kw_only=True)
    properties = attrs.field(type=List[SegmentPropertyBase], default=[], kw_only=True)

    def __len__(self):
        return len(self.ids)

    def __attrs_post_init__(self):
        if len(self.properties) > 0:
            _validate_properties(self.ids, self.properties)


@attrs.define
class LabelProperty(SegmentPropertyBase):
    id = attrs.field(default="label", type=str, converter=str, kw_only=True)
    type = attrs.field(init=False, default="label", type=str)
    values = attrs.field(type=list, converter=list_of_strings, kw_only=True)
    description = attrs.field(type=Optional[str], default=None, kw_only=True)

    def __len__(self):
        return len(self.values)


@attrs.define
class DescriptionProperty(SegmentPropertyBase):
    id = attrs.field(default="description", type=str, converter=str, kw_only=True)
    type = attrs.field(init=False, default="description", type=str)
    values = attrs.field(type=list, converter=list_of_strings, kw_only=True)
    description = attrs.field(type=Optional[str], default=None, kw_only=True)

    def __len__(self):
        return len(self.values)


@attrs.define
class StringProperty(SegmentPropertyBase):
    id = attrs.field(type=str, converter=str, kw_only=True)
    type = attrs.field(init=False, default="string", type=str)
    values = attrs.field(type=list, converter=list_of_strings, kw_only=True)
    description = attrs.field(type=Optional[str], default=None, kw_only=True)

    def __len__(self):
        return len(self.values)


@attrs.define
class NumberProperty(SegmentPropertyBase):
    id = attrs.field(type=str, converter=str, kw_only=True)
    type = attrs.field(init=False, default="number", type=str)
    values = attrs.field(type=list, kw_only=True)
    description = attrs.field(type=Optional[str], default=None, kw_only=True)
    data_type = attrs.field(
        type=str,
        kw_only=True,
        validator=attrs.validators.in_(ALLOWED_NUMBER_DATA_TYPES),
    )

    def __attrs_post_init__(self):
        prop_dtype = np.dtype(self.data_type)
        self.values = (
            np.array(self.values).astype(prop_dtype, casting="same_kind").tolist()
        )

    def __len__(self):
        return len(self.values)


@attrs.define
class TagProperty(SegmentPropertyBase):
    id = attrs.field(type=str, kw_only=True, default="tags")
    type = attrs.field(init=False, type=str, default="tags")
    tags = attrs.field(
        type=List[str],
        converter=space_to_underscore,
        validator=attrs.validators.not_(attrs.validators.matches_re(r"^\#")),
        kw_only=True,
    )
    tag_descriptions = attrs.field(
        type=List[str],
        default=None,
        kw_only=True,
    )
    values = attrs.field(
        type=List[List[int]],
        converter=sort_tag_arrays,
        kw_only=True,
    )

    def __len__(self):
        return len(self.values)

    def __repr__(self):
        return f"TagProperty(id='{self.id}', {len(self.tags)} tags, {len(self.values)} values)"


def prop_filter(attr, value):
    "Filters out None for optional attributes for use in 'attrs.asdict' conversion"
    return value is not None or attr.name not in NONE_ATTR_FILTERS


prop_to_dict = partial(attrs.asdict, filter=prop_filter)


def _validate_properties(
    ids,
    properties,
):
    n_ids = len(ids)
    for prop in properties:
        if len(prop) != n_ids:
            msg = f"Property {prop} does not have the same number of entries as the id list."
            raise ValueError(msg)


def build_segment_properties(
    ids: List[int],
    properties: list,
):
    return {
        "@type": "neuroglancer_segment_properties",
        "inline": prop_to_dict(InlineProperties(ids=ids, properties=properties)),
    }


def _find_column_dtype(column):
    "Get data type string from a dataframe column"
    if column.dtype in ALLOWED_NUMBER_DATA_TYPES:
        return str(column.dtype)
    elif column.dtype == "int64":
        return "int32"
    elif column.dtype == "float64":
        return "float32"
    elif column.dtype == "object":
        try:
            column.astype("float32")
        except Exception:
            raise ValueError(
                f"Column {column} has an unsupported data type {column.dtype}"
            )
    else:
        raise ValueError(f"Column {column} has an unsupported data type {column.dtype}")


def _tag_descriptions(tags, tag_descriptions):
    if tag_descriptions is None:
        return None
    else:
        return [tag_descriptions.get(tag, tag) for tag in tags]


def _make_tag_map(
    df, value_columns, bool_columns, allow_disambiguation, prepend_col_name
):
    if prepend_col_name:
        for col in value_columns:
            df[col] = df[col].apply(lambda x: f"{col}:{x}")
    unique_tags = {}
    for col in value_columns:
        col_tags = df[col].unique()
        if isinstance(col_tags, np.ndarray):
            unique_tags[col] = sorted(
                [x for x in col_tags.tolist() if not is_null_value(x)]
            )
        else:
            unique_tags[col] = [
                x for x in col_tags.sort_values().tolist() if not is_null_value(x)
            ]
    for col in bool_columns:
        unique_tags[col] = [col]
    vals, counts = np.unique(
        np.concatenate([v for v in unique_tags.values()]), return_counts=True
    )
    duplicates = vals[counts > 1]
    swap_values = []
    if len(duplicates) > 0:
        if not allow_disambiguation:
            raise ValueError(f"Duplicate tags found: {duplicates}")
        for col in unique_tags:
            for dup in duplicates:
                if dup in unique_tags[col]:
                    if col in value_columns:
                        df.loc[:, col] = df[col].replace(dup, f"{col}:{dup}")
                        swap_values.append(
                            (col, unique_tags[col].index(dup), f"{col}:{dup}")
                        )
    for col, idx, swap in swap_values:
        unique_tags[col][idx] = swap
    tags = np.concatenate([v for v in unique_tags.values()]).tolist()
    return tags, {tag: i for i, tag in enumerate(tags)}, df


def _generate_tag_values(df, value_columns, bool_columns, tag_map):
    index_col = "new_column_index_temp_"
    tag_df = df.assign(**{index_col: np.arange(len(df))})
    concat_dfs = []
    if len(value_columns) > 0:
        tag_df_long = (
            tag_df.melt(
                id_vars=index_col,
                value_vars=value_columns,
            )
            .dropna(how="any")
            .sort_values(by=index_col)
        )
        tag_df_long["tag_value"] = (
            tag_df_long["value"].astype(object).apply(lambda x: tag_map[x]).values
        )
        concat_dfs.append(tag_df_long[[index_col, "tag_value"]])
    if len(bool_columns) > 0:
        for col in bool_columns:
            concat_dfs.append(
                pd.DataFrame(
                    {
                        index_col: np.flatnonzero(tag_df[col]),
                        "tag_value": tag_map[col],
                    }
                )
            )
    tag_values = (
        pd.concat(concat_dfs, ignore_index=True)
        .groupby(index_col)["tag_value"]
        .apply(lambda x: sorted(list(x)))
    )
    # IDs without a tag would be missing from the above
    index_missing = np.setdiff1d(
        np.arange(len(tag_df)),
        tag_values.index,
    )
    for idx in index_missing:
        tag_values[idx] = []
    return tag_values.sort_index().tolist()


def _make_tag_property(
    df,
    value_columns,
    bool_columns,
    tag_descriptions,
    name="tags",
    allow_disambiguation=True,
    prepend_col_name=False,
):
    if value_columns is None:
        value_columns = []
    if bool_columns is None:
        bool_columns = []
    tags, tag_map, tag_df = _make_tag_map(
        df[value_columns + bool_columns].replace({"": None}).copy(),
        value_columns,
        bool_columns,
        allow_disambiguation,
        prepend_col_name,
    )
    tag_values = _generate_tag_values(tag_df, value_columns, bool_columns, tag_map)
    return TagProperty(
        id=name,
        tags=tags,
        values=tag_values,
        tag_descriptions=_tag_descriptions(tags, tag_descriptions),
    )


class SegmentProperties:
    def __init__(
        self,
        ids: List[int],
        label_property: Optional[LabelProperty] = None,
        description_property: Optional[DescriptionProperty] = None,
        tag_properties: Optional[TagProperty] = None,
        string_properties: Optional[Union[StringProperty, List[StringProperty]]] = None,
        number_properties: Optional[Union[NumberProperty, List[NumberProperty]]] = None,
    ):
        self.ids = ids
        self.label_property = label_property
        self.description_property = description_property
        self.tag_properties = tag_properties

        if isinstance(string_properties, StringProperty):
            string_properties = [string_properties]
        self.string_properties = string_properties

        if isinstance(number_properties, NumberProperty):
            number_properties = [number_properties]
        self.number_properties = number_properties

    def __len__(self):
        return len(self.ids)

    def _property_list(self):
        single_prop_list = [
            prop
            for prop in [
                self.label_property,
                self.description_property,
                self.tag_properties,
            ]
            if prop is not None
        ]
        multi_prop_list = list(
            chain.from_iterable(
                [
                    prop
                    for prop in [
                        self.string_properties,
                        self.number_properties,
                    ]
                    if prop is not None
                ]
            )
        )
        return single_prop_list + multi_prop_list

    def __repr__(self):
        return f"SegmentProperties ({len(self.ids)} segments, {len(self._property_list())} properties)"

    def __str__(self):
        return self.__repr__()

    def property_description(self):
        return [(prop.id, prop.type, len(prop)) for prop in self._property_list()]

    def to_dict(self):
        "Converts the segment properties to a dictionary for use in neuroglancer"
        return build_segment_properties(
            self.ids,
            self._property_list(),
        )

    def to_dataframe(self):
        "Converts the segment properties to a pandas dataframe"
        df_dict = {"ids": self.ids}
        for prop in self._property_list():
            if (
                isinstance(prop, (DescriptionProperty, LabelProperty, NumberProperty, StringProperty))
            ):
                df_dict[prop.id] = prop.values
            elif isinstance(prop, TagProperty):
                for ii, tag in enumerate(prop.tags):
                    df_dict[tag] = [ii in tags for tags in prop.values]
        return pd.DataFrame(df_dict)

    @classmethod
    def from_dataframe(
        cls,
        df: pd.DataFrame,
        id_col: str = "pt_root_id",
        label_col: Optional[Union[str, List[str]]] = None,
        description_col: Optional[str] = None,
        string_cols: Optional[Union[str, List[str]]] = None,
        number_cols: Optional[Union[str, List[str]]] = None,
        tag_value_cols: Optional[Union[str, List[str]]] = None,
        tag_bool_cols: Optional[List[str]] = None,
        tag_descriptions: Optional[dict] = None,
        allow_disambiguation: bool = True,
        label_separator: str = "_",
        label_format_map: Optional[str] = None,
        prepend_col_name: bool = False,
    ):
        """Generate a segment property object from a pandas dataframe based on column

        Parameters
        ----------
        df : pd.DataFrame
            Dataframe containing propeties
        id_col : str, optional
            Name of the column with object ids, by default "pt_root_id"
        label_col : Optional[str, list[str]], optional
            Name of column or columns to use for producing labels, by default None.
            If multiple columns are provided, they will be concatenated with the label_separator.
            Null values are skipped.
        description_col : Optional[str], optional
            Name of column to use for producing descriptions, by default None
        string_cols : Optional[Union[str, list[str]]], optional
            Column (or list of columns) to use for string properties, by default None.
        number_cols : Optional[Union[str, list[str]]], optional
            Column (or list of columns) to use for numeric properties, by default None.
        tag_value_cols : Optional[list[str]], optional
            Column (or list of columns) to generate tags based on unique values.
            Each column produces one tag per row based on the value, by default None
        tag_bool_cols : Optional[list[str]], optional
            List of columns to generate tags based on boolean values where each column is a tag, and each id gets the tag if it has a True in its row.
            By default None.
        tag_descriptions : Optional[dict], optional
            Dictionary of tag values to long-form tag descriptions, by default None.
            Tags without a key/value are passed through directly.
        allow_disambiguation : bool, optional
            If True, will prepend the column name in the case of duplicate tags, by default True.
        label_separator : str, optional
            Separator to use when assembling multiple columns into a label, by default "_"
        label_format_map : Optional[str], optional
            Format string to use for label formatting, by default None.
            If provided, will override the label separator and use the format string to format the label
            via the "format" function, replacing the column names in `{..}` with the values.
            For example, "{cell_class}: {cell_type}_{region}" would pluck values from the columns
            "cell_class", "cell_type", and "region". Label columns will be ignored and the format string is not validated.
        prepend_col_name : bool, optional
            If True, will prepend the column name to tag values, by default False.
            This will effectively disambiguate all tags as well.

        Returns
        -------
        SegmentProperties
            Segment properties object
        """
        ids = df[id_col].tolist()
        properties = {}
        if label_col or label_format_map:
            if isinstance(label_col, str):
                label_col = [label_col]
            if label_format_map:
                label_vals = df.apply(
                    lambda x: label_format_map.format(**x.to_dict()), axis=1
                ).to_list()
            else:
                label_vals = [
                    label_separator.join(filter(None, r))
                    for r in df[label_col]
                    .where(pd.notnull(df[label_col]), "")
                    .astype(str)
                    .values
                ]
            properties["label_property"] = LabelProperty(values=label_vals)
        if description_col:
            properties["description_property"] = DescriptionProperty(
                values=df[description_col].tolist()
            )
        if string_cols:
            if isinstance(string_cols, str):
                string_cols = [string_cols]
            properties["string_properties"] = [
                StringProperty(id=col, values=df[col].tolist()) for col in string_cols
            ]
        if number_cols:
            if isinstance(number_cols, str):
                number_cols = [number_cols]
            properties["number_properties"] = [
                NumberProperty(
                    id=col,
                    values=df[col].tolist(),
                    data_type=_find_column_dtype(df[col]),
                )
                for col in number_cols
            ]
        if tag_value_cols or tag_bool_cols:
            if isinstance(tag_value_cols, str):
                tag_value_cols = [tag_value_cols]
            if isinstance(tag_bool_cols, str):
                tag_bool_cols = [tag_bool_cols]
            properties["tag_properties"] = _make_tag_property(
                df,
                tag_value_cols,
                tag_bool_cols,
                tag_descriptions,
                allow_disambiguation=allow_disambiguation,
                prepend_col_name=prepend_col_name,
            )
        return cls(ids, **properties)

    @classmethod
    def from_dict(
        cls,
        seg_prop_dict: dict,
    ):
        """Generate a segment property object from a segment property dictionary

        Parameters
        ----------
        seg_prop_dict : dict
            Segment property dictionary, as imported from the json.

        Returns
        -------
        SegmentProperties
            Segment properties object
        """
        ids = seg_prop_dict["inline"]["ids"]
        props = seg_prop_dict["inline"]["properties"]
        prop_classes = {}
        for prop in props:
            if prop["type"] == "label":
                prop_classes["label_property"] = LabelProperty(
                    values=prop["values"],
                    description=prop.get("description"),
                )
            elif prop["type"] == "description":
                prop_classes["description_property"] = DescriptionProperty(
                    values=prop["values"],
                    description=prop.get("description"),
                )
            elif prop["type"] == "string":
                if "string_properties" not in prop_classes:
                    prop_classes["string_properties"] = []
                prop_classes["string_properties"].append(
                    StringProperty(
                        id=prop["id"],
                        values=prop["values"],
                        description=prop.get("description"),
                    )
                )
            elif prop["type"] == "number":
                if "number_properties" not in prop_classes:
                    prop_classes["number_properties"] = []
                prop_classes["number_properties"].append(
                    NumberProperty(
                        id=prop["id"],
                        values=prop["values"],
                        data_type=prop["data_type"],
                        description=prop.get("description"),
                    )
                )
            elif prop["type"] == "tags":
                prop_classes["tag_properties"] = TagProperty(
                    tags=prop["tags"],
                    values=prop["values"],
                    tag_descriptions=prop.get("tag_descriptions"),
                )
        return cls(ids, **prop_classes)



import logging
import os
from typing import Literal, Optional

import attrs
import caveclient

logger = logging.getLogger(__name__)

MAINLINE_NAMES = ["spelunker", "mainline", "cave-explorer"]
SEUNGLAB_NAMES = ["seunglab"]
default_seunglab_neuroglancer_base = "https://neuromancer-seung-import.appspot.com/"
default_mainline_neuroglancer_base = "https://spelunker.cave-explorer.org/"

DEFAULT_TARGET_SITE = os.environ.get("NGLUI_DEFAULT_TARGET_SITE", "seunglab")
DEFAULT_URL = os.environ.get("NGLUI_DEFAULT_TARGET_URL", None)
DEFAULT_DATASTACK_NAME = os.environ.get("NGLUI_DEFAULT_DATASTACK_NAME", None)

__all__ = [
    "is_mainline",
    "is_seunglab",
    "check_target_site",
    "set_default_config",
    "get_default_config",
    "reset_default_config",
    MAINLINE_NAMES,
    SEUNGLAB_NAMES,
]


def is_mainline(target_name):
    return target_name in MAINLINE_NAMES


def is_seunglab(target_name):
    return target_name in SEUNGLAB_NAMES


def check_target_site(ngl_url, client):
    """
    Check neuroglancer info to determine which kind of site a neuroglancer URL is.
    """
    ngl_info = client.state.get_neuroglancer_info(ngl_url)
    if len(ngl_info) == 0:
        return SEUNGLAB_NAMES[0]
    else:
        return MAINLINE_NAMES[0]


@attrs.define
class NGLUIConfig:
    target_site = attrs.field(type=str)
    target_url = attrs.field(type=str)
    seunglab_fallback_url = attrs.field(
        default=default_seunglab_neuroglancer_base, type=str
    )
    mainline_fallback_url = attrs.field(
        default=default_mainline_neuroglancer_base, type=str
    )
    caveclient = attrs.field(default=None)
    datastack_name = attrs.field(default=DEFAULT_DATASTACK_NAME, type=str)

    def __attrs_post_init__(self):
        # validate target site
        if (
            self.target_site not in MAINLINE_NAMES + SEUNGLAB_NAMES
            and self.target_site is not None
        ):
            raise ValueError(
                f"target_site must be one of {MAINLINE_NAMES + SEUNGLAB_NAMES}"
            )
        # Set and validate caveclient if info is provided
        if self.caveclient is None and self.datastack_name is not None:
            try:
                self.caveclient = caveclient.CAVEclient(self.datastack_name)
            except Exception as e:
                msg = f"Could not create CAVEclient with datastack_name {self.datastack_name}: {e}"
                logger.error(msg)
        if self.caveclient is not None and self.datastack_name is None:
            self.datastack_name = self.caveclient.datastack_name
        if self.datastack_name is not None and self.caveclient is not None:
            if self.caveclient.datastack_name != self.datastack_name:
                raise ValueError(
                    f"CAVEclient datastack_name {self.caveclient.datastack_name} does not match provided datastack_name {self.datastack_name}"
                )

        # Set and validate target_url if target_site is provided
        if self.target_url is not None and self.target_site is None:
            self.target_site = check_target_site(self.target_url, self.caveclient)
        if self.target_site is not None and self.target_url is None:
            if is_seunglab(self.target_site):
                self.target_url = self.seunglab_fallback_url
            else:
                self.target_url = self.mainline_fallback_url
        if self.target_site is not None and self.target_url is not None:
            if self.caveclient is not None:
                url_based_site = check_target_site(self.target_url, self.caveclient)
                if is_mainline(url_based_site) != is_mainline(self.target_site):
                    logging.warning(
                        f"target_site {self.target_site} does not match target_url {self.target_url}. Assuming manual setting is correct."
                    )


default_config = NGLUIConfig(
    target_site=DEFAULT_TARGET_SITE,
    target_url=DEFAULT_URL,
)
default_key = "default"
NGL_CONFIG = {default_key: default_config}  # type: dict[str, NGLUIConfig]


def get_default_config(config_key: str = None):
    """Get the current configuration for nglui viewers and statebuilders.

    Parameters
    ----------
    config_key : str, optional
        Key for the configuration setting, by default "default"
    """
    if config_key is None:
        config_key = default_key
    return attrs.asdict(NGL_CONFIG[config_key])


def set_default_config(
    target_site: Optional[
        Literal["seunglab", "mainline", "cave-explorer", "spelunker"]
    ] = None,
    target_url: str = None,
    seunglab_fallback_url: str = None,
    mainline_fallback_url: str = None,
    datastack_name: str = None,
    caveclient: "caveclient.CAVEclient" = None,
    url_from_client: bool = False,
    config_key: str = "default",
):
    """Set default configuration for nglui viewers and statebuilders.

    Parameters
    ----------
    target_site : str, optional
        Target site for the neuroglancer viewer, by default None
    target_url : str, optional
        Target URL for the neuroglancer viewer, by default None
    seunglab_url : str, optional
        URL for the seunglab neuroglancer viewer, by default None
    mainline_url : str, optional
        URL for the mainline neuroglancer viewer, by default None
    datastack_name : str, optional
        Name of the datastack, by default None
    caveclient : caveclient.CAVEclient, optional
        CAVEclient object, by default None.
    """
    if config_key in NGL_CONFIG:
        curr_config = get_default_config(config_key)
    else:
        curr_config = attrs.asdict(default_config)
    if target_site is not None or target_url is not None:
        curr_config["target_site"] = target_site
        curr_config["target_url"] = target_url
    if seunglab_fallback_url is not None:
        curr_config["seunglab_fallback_url"] = seunglab_fallback_url
    if mainline_fallback_url is not None:
        curr_config["mainline_fallback_url"] = mainline_fallback_url
    if caveclient is not None or datastack_name is not None:
        curr_config["caveclient"] = caveclient
        curr_config["datastack_name"] = datastack_name
        if url_from_client:
            curr_config["target_site"] = None
            curr_config["target_url"] = caveclient.info.viewer_site()
    NGL_CONFIG[config_key] = NGLUIConfig(**curr_config)


def reset_config(
    config_key: str = "default",
):
    """Reset the configuration for nglui viewers and statebuilders to the default.

    Parameters
    ----------
    config_key : str, optional
        Key for the configuration setting, by default "default"
    """
    NGL_CONFIG[config_key] = default_config


def neuroglancer_url(
    url: Optional[str] = None,
    target_site: Optional[str] = None,
    config_key: str = "default",
):
    """
    Check neuroglancer info to determine which kind of site a neuroglancer URL is.
    If either url or target_site are provided, it will use these values, looking up target site
    from the fallback values in the config. Otherwise, it falls back to the value of
    "target_url" in the config.

    Parameters
    ----------
    url : str, optional
        URL to check, by default None
    target_site : str, optional
        Target site to check, by default None

    Returns
    -------
    str
        URL of the neuroglancer viewer
    """
    if url is not None:
        return url
    if target_site is None:
        url = get_default_config(config_key)["target_url"]
    elif is_seunglab(target_site):
        url = get_default_config(config_key)["seunglab_fallback_url"]
    else:
        url = get_default_config(config_key)["mainline_fallback_url"]
    return url


def reset_default_config():
    NGL_CONFIG[default_key] = default_config
